from .asymptotic import (
    asymptotic_H12_Retheta,
    asymptotic_H32_Retheta,
)
from .bl import (
    calc_d1,
    calc_d2,
    calc_d3,
    calc_d99_1d,
    calc_dRC,
    calc_I2,
    calc_I3,
    calc_profile_edge_1d,
    calc_wake_parameter_1d,
)
from .blasius import Blasius_solution
from .cgd import cgd
from .composite_profile import composite_profile_CMN2009
from .compressible_transform import (
    comp_transform_GFM,
    comp_transform_TL,
    comp_transform_VD,
    comp_transform_VIPL,
)
from .confidence_interval import calc_var_bmbc, confidence_interval_unbiased
from .eas3 import eas3
from .eas4 import eas4
from .fig_ax_constructor import fig_ax_grid
from .freestream_parameters import freestream_parameters
from .gradient import fd_coeff_calculator, gradient
from .h5 import h5_print_contents
from .rgd import rgd
from .set_mpl_env import set_mpl_env
from .spd import spd
from .utils import even_print, format_nbytes, format_time_string, step
from .ztmd import ztmd
from .ztmd_loader import ztmd_loader

__all__ = [
    'Blasius_solution',
    'asymptotic_H12_Retheta',
    'asymptotic_H32_Retheta',
    'calc_I2',
    'calc_I3',
    'calc_d1',
    'calc_d2',
    'calc_d3',
    'calc_d99_1d',
    'calc_dRC',
    'calc_profile_edge_1d',
    'calc_var_bmbc',
    'calc_wake_parameter_1d',
    'cgd',
    'comp_transform_GFM',
    'comp_transform_TL',
    'comp_transform_VD',
    'comp_transform_VIPL',
    'composite_profile_CMN2009',
    'confidence_interval_unbiased',
    'eas3',
    'eas4',
    'even_print',
    'fd_coeff_calculator',
    'fig_ax_grid',
    'format_nbytes',
    'format_time_string',
    'freestream_parameters',
    'gradient',
    'h5_print_contents',
    'rgd',
    'set_mpl_env',
    'spd',
    'step',
    'ztmd',
    'ztmd_loader',
    ]


import numpy as np
import scipy as sp

'''
========================================================================
Composite profile of Chauhan et al. (2009)
https://doi.org/10.1088/0169-5983/41/2/021404
========================================================================
'''

# ======================================================================

class composite_profile_CMN2009():
    '''
    Tool for calculating composite profile described in Chauhan et al. (2009)
    'Criteria for assessing experiments in zero pressure gradient boundary layers'
    Chauhan, Monkewitz and Nagib
    https://doi.org/10.1088/0169-5983/41/2/021404
    '''
    
    def __init__(self, Re_tau, k=0.384, Pi=0.55, a=None, C=None, a2=132.8410, a3=-166.2041, a4=71.9114):
        '''
        k is the Von Kármán constant κ (asymptotic ≈0.384)
        Reτ is the friction Reynolds number δ+ = δ/δν = δ·uτ/ν
        C is the log law semilog constant (asymptotic ≈4.127, sometimes 5.0-5.2 in classic literature)
        a is a parameter directly related to C, see paper
        Π is the Coles wake parameter
        -----
        a = -10.3061 corresponds to C = 4.17
        '''
        self.k = k
        self.Re_tau = Re_tau
        self.Pi = Pi
        self.a2 = a2
        self.a3 = a3
        self.a4 = a4
        
        ## assert that exacly one of a or C was set
        if (a is None) and (C is None):
            raise ValueError('set one of a or C')
        if (a is not None) and (C is not None):
            raise ValueError('set either a or C, not both')
        
        if (C is not None): ## find 'a' given C
            
            self.C = C
            
            if not isinstance(C,float):
                raise ValueError
            
            def __f_opti(a,C):
                y_plus_inf = 1e12
                B_asymptotic = (1/k)*np.log((y_plus_inf-a)/-a) + self.__uplus_inner_B(y_plus_inf,a,k) - (1/k)*np.log(y_plus_inf)
                root = np.abs(B_asymptotic-C)
                return root
            
            sol = sp.optimize.least_squares(fun=__f_opti,
                                            args=(C,),
                                            x0=-10.,
                                            xtol=1e-15,
                                            ftol=1e-15,
                                            gtol=1e-15,
                                            method='dogbox',
                                            bounds=(-100.,+100.))
            if not sol.success:
                raise ValueError
            
            self.a = float(sol.x[0])
        
        else:
            
            self.a = a
            
            def __f_opti(C,a):
                y_plus_inf = 1e12
                B_asymptotic = (1/k)*np.log((y_plus_inf-a)/-a) + self.__uplus_inner_B(y_plus_inf,a,k) - (1/k)*np.log(y_plus_inf)
                root = np.abs(B_asymptotic-C)
                return root
            
            sol = sp.optimize.least_squares(fun=__f_opti,
                                            args=(a,),
                                            x0=-10.,
                                            xtol=1e-15,
                                            ftol=1e-15,
                                            gtol=1e-15,
                                            method='dogbox',
                                            bounds=(-100.,+100.))
            if not sol.success:
                raise ValueError
            
            self.C = float(sol.x[0])
    
    def get_uplus_inner(self,y_plus):
        '''
        Calculate inner U+ profile
        '''
        
        a = self.a ## related to B for log law
        k = self.k
        
        ## this B is a function of y+ for U+_inner i.e. is a vector
        B = self.__uplus_inner_B(y_plus,a,k)
        
        ## the 'inner' U+ profile
        u_plus_inner = np.copy( (1/k)*np.log((y_plus-a)/-a) + B )
        
        ## log-law constant is [U+] - (1/k)ln(y+) for [y+ -> inf]
        y_plus_inf = 1e14
        B_asymptotic = (1/k)*np.log((y_plus_inf-a)/-a) + self.__uplus_inner_B(y_plus_inf,a,k) - (1/k)*np.log(y_plus_inf)
        self.B = B_asymptotic
        #print(f'B={B_asymptotic:0.14f}')
        
        return u_plus_inner
    
    def __uplus_inner_B(self,y_plus,a,k):
        
        alpha = (-1/k - a)/2
        beta  = np.sqrt(-2*a*alpha - alpha**2)
        R     = np.sqrt( alpha**2 + beta**2 )
        
        #T1 = (4*alpha+a) * np.log( -(a/R)*np.sqrt( (y_plus-alpha)**2 + beta**2 ) / ( y_plus - a ) ) \
        #     + (alpha/beta)*(4*alpha + 5*a)*( np.arctan((y_plus-alpha)/beta) + np.arctan(alpha/beta) )
        T1 = (4*alpha+a) * np.log( -(a/R)*np.sqrt( (y_plus-alpha)**2 + beta**2 ) / ( y_plus - a ) ) \
             + (alpha/beta)*(4*alpha + 5*a)*( np.arctan2((y_plus-alpha),beta) + np.arctan2(alpha,beta) )
        
        T2 = R**2 / (a*(4*alpha-a)) * T1
        
        return T2
    
    def f_wake_Chauhan_exp(self,eta,Pi,a2,a3,a4):
        '''
        Exponential wake function W from Chauhan 2009
        '''
        T1 = 1. - np.exp( -(1/4)*( 5*a2 + 6*a3 + 7*a4 )*eta**4  + a2*eta**5 + a3*eta**6 + a4*eta**7 )
        T2 = 1. - np.exp( -(a2 + 2*a3 + 3*a4)/4 )
        T3 = 1. - 1/(2*Pi) * np.log(eta)
        return (T1/T2) * T3
    
    def get_wake(self, y_ov_delta, Pi):
        '''
        See Chauhan 2009
        '''
        
        a2 = self.a2
        a3 = self.a3
        a4 = self.a4
        
        if (y_ov_delta[0]!=0.):
            raise ValueError('y_ov_delta should start at 0')
        W = np.zeros_like(y_ov_delta)
        W[:] = np.nan
        i_valid = np.where( (y_ov_delta>0.) & (y_ov_delta<=1.) )
        eta = np.copy(y_ov_delta[i_valid])
        W[i_valid] = self.f_wake_Chauhan_exp(eta,Pi,a2,a3,a4)
        W[np.where(y_ov_delta==0.)] = 0.
        #if np.isnan(W).any():
        #    raise ValueError
        return W
    
    def get_integral_quantities(self, y_plus, u_plus, interp_kind='cubic'):
        '''
        δ*=δ1, θ=δ2, Reθ, Reδ1, Reδ2, H
        '''
        
        if not isinstance(y_plus, np.ndarray):
            raise ValueError
        if not isinstance(u_plus, np.ndarray):
            raise ValueError
        if (y_plus.ndim!=1):
            raise ValueError('y_plus.ndim!=1')
        if (u_plus.ndim!=1):
            raise ValueError('u_plus.ndim!=1')
        if (u_plus.shape != y_plus.shape):
            raise ValueError
        
        ## y_max < δ
        if (y_plus.max() < self.Re_tau):
            raise ValueError('y+_max of profile cannot be <Reτ --> y_max < δ')
        
        ## these are simply used locally to dimensionalize
        nu = 1e-5 ## [m^2/s]
        d = 37. ## [m]
        
        u_tau = nu * self.Re_tau / d ## [m/s]
        
        #sc_l_in = nu / u_tau
        sc_l_in = d / self.Re_tau
        
        U_inf_plus = self.get_uplus_inner(y_plus=self.Re_tau) + (2*self.Pi/self.k)
        U_inf = U_inf_plus * u_tau ## [m/s]
        
        u = np.copy( u_plus * u_tau ) ## [m/s]
        y = np.copy( y_plus * sc_l_in ) ## [m]
        
        integrand_theta = (u/U_inf)*(1-(u/U_inf))
        integrand_dstar = 1-(u/U_inf)
        
        #print(y.max())
        #print(d)
        #if not np.isclose(y.max(), d, rtol=1e-6):
        #    raise AssertionError
        
        theta_     = sp.integrate.cumulative_trapezoid(y=integrand_theta, x=y, initial=0.)
        theta_func = sp.interpolate.interp1d(y, theta_, kind=interp_kind, bounds_error=False, fill_value='extrapolate')
        theta      = theta_func(d)
        
        dstar_     = sp.integrate.cumulative_trapezoid(y=integrand_dstar, x=y, initial=0.)
        dstar_func = sp.interpolate.interp1d(y, dstar_, kind=interp_kind, bounds_error=False, fill_value='extrapolate')
        dstar      = dstar_func(d)
        
        Re_theta = theta * U_inf / nu
        Re_dstar = dstar * U_inf / nu
        H        = dstar / theta
        
        #dRC = U_inf_plus / dstar ## Δ=U_inf+/δ
        
        return Re_theta, Re_dstar, H
    
    def calc_uplus_composite(self,y_plus):
        
        y_ov_delta = np.copy( y_plus / self.Re_tau )
        i_gt_1 = np.where(y_ov_delta>1.)
        
        u_plus_inner = self.get_uplus_inner(y_plus)
        W = self.get_wake(y_ov_delta, Pi=self.Pi)
        i_nan = np.where(np.isnan(W))
        
        np.testing.assert_array_equal(i_gt_1,i_nan)
        
        ## U+_composite = U+_inner + U+_wake
        u_plus = np.copy( u_plus_inner + (2*self.Pi/self.k)*W )
        
        ## for y>δ, U+ = U+_inf
        U_inf_plus = self.get_uplus_inner(y_plus=self.Re_tau) + (2*self.Pi/self.k)
        u_plus[i_gt_1] = U_inf_plus
        
        return u_plus
    
    def __call__(self,y_plus):
        '''
        Calling object returns u+(y+)
        '''
        return self.calc_uplus_composite(y_plus)



import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import PchipInterpolator, interp1d
from scipy.optimize import least_squares, root_scalar

'''
========================================================================
1D boundary layer profile kernels
========================================================================
'''

def _check_validity_yx(y,x):
    '''
    Check 1D coordinate [y] and generic 1D variable [x]
    '''
    if not isinstance(y,np.ndarray):
        raise ValueError('y must be type np.ndarray')
    if not isinstance(x,np.ndarray):
        raise ValueError('x must be type np.ndarray')
    if y.ndim != 1 or x.ndim != 1:
        raise ValueError('.ndim != 1')
    if y.shape[0] != x.shape[0]:
        raise ValueError('y,x shapes not equal')
    if not np.all(np.diff(y) > 0): ## check (+) monotonicity
        raise ValueError("y must be monotonically increasing")
    if y[0] < -1e-4:
        raise ValueError("y[0] < 0")
    if not np.isclose(y[0], 0., atol=1e-4):
        raise ValueError("y[0] not sufficiently close to 0")
    return

def _check_validity_jn(j_edge,n):
    '''
    Check edge index j_edge with respect to 1D shape n
    '''
    if not isinstance(j_edge,(int,np.integer)):
        raise ValueError("j_edge must be int")
    if j_edge>=n:
        raise ValueError("j_edge>=n")
    if j_edge<-n:
        raise ValueError("j_edge<-n")
    if j_edge==0:
        raise RuntimeError('j_edge==0')
    return

def _determine_ju_edge(y,u,j_edge,u_edge,**kwargs):
    '''
    Helper function to handle the various
      j_edge,u_edge arg-passing patterns
    '''
    n     = y.shape[0]
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    ## (1/4) Both j_edge & u_edge provided: just check
    if j_edge is not None and u_edge is not None:
        if check:
            np.testing.assert_allclose(u[j_edge], u_edge, rtol=rtol)
    
    ## (2/4) Neither j_edge nor u_edge provided: take max(u)
    elif j_edge is None and u_edge is None:
        u_edge = u.max()
        j_edge = int(np.argmax(u))
        np.testing.assert_allclose(u[j_edge], u_edge, rtol=1e-6)
    
    ## (3/4) j_edge passed but no u_edge
    elif j_edge is not None and u_edge is None:
        u_edge = float(u[j_edge])
    
    ## (4/4) u_edge given but no j_edge
    elif j_edge is None and u_edge is not None:
        jj = np.where( u >= u_edge )[0] ## Freestream indices
        if jj.shape[0]==0: ## No index is > u_edge
            j_edge = n-1 ## Index of last element
        else:
            j_edge = int(jj.min()) ## Lowest 'freestream' index
    
    else:
        raise ValueError
    
    return j_edge, u_edge

# ======================================================================

def calc_d1(y, u, rho=None, j_edge=None, u_edge=None, rho_edge=None, **kwargs):
    '''
    Displacement (mass-flux deficit) thickness δ1 = δ*
    -----
    δ1 = ∫ ( 1 - ρu / (ρe·ue) ) dy
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    for val in (u_edge, rho_edge):
        if val is not None and not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    
    ## If ρ_edge is passed, then ρ must also be passed
    if rho_edge is not None and rho is None:
        raise ValueError('ρ_edge passed but ρ not passed')
    
    if rho is None: ## no ρ passed, set ρ=const=1 (incomp.)
        rho = np.ones_like(u)
    rho = np.asarray(rho)
    
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    if rho_edge is None:
        rho_edge = rho[j_edge]
    
    integrand = 1. - (rho*u)/(rho_edge*u_edge) ## δ1 integrand
    d1 = cumulative_trapezoid(y=integrand, x=y, initial=0.)[j_edge] ## δ1
    return d1

def calc_d2(y, u, rho=None, j_edge=None, u_edge=None, rho_edge=None, **kwargs):
    '''
    Momentum deficit thickness δ2 = θ
    -----
    δ2 = ∫ ( ρu / (ρe·ue) ) ( 1 - u / ue ) dy
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    for val in (u_edge, rho_edge):
        if val is not None and not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    
    ## If ρ_edge is passed, then ρ must also be passed
    if rho_edge is not None and rho is None:
        raise ValueError('ρ_edge passed but ρ not passed')
    
    if rho is None: ## no ρ passed, set ρ=const=1 (incomp.)
        rho = np.ones_like(u)
    rho = np.asarray(rho)
    
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    if rho_edge is None:
        rho_edge = rho[j_edge]
    
    integrand = (rho*u)/(rho_edge*u_edge) * (1. - u/u_edge) ## δ2 integrand
    d2 = cumulative_trapezoid(y=integrand, x=y, initial=0.)[j_edge] ## δ2
    return d2

def calc_d3(y, u, rho=None, j_edge=None, u_edge=None, rho_edge=None, **kwargs):
    '''
    Kinetic energy deficit thickness δ3
    -----
    δ3 = ∫ ( ρu / (ρe·ue) ) ( 1 - (u / ue)^2 ) dy
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    for val in (u_edge, rho_edge):
        if val is not None and not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    
    ## If ρ_edge is passed, then ρ must also be passed
    if rho_edge is not None and rho is None:
        raise ValueError('ρ_edge passed but ρ not passed')
    
    if rho is None: ## no ρ passed, set ρ=const=1 (incomp.)
        rho = np.ones_like(u)
    rho = np.asarray(rho)
    
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    if rho_edge is None:
        rho_edge = rho[j_edge]
    
    integrand = (rho*u)/(rho_edge*u_edge) * (1. - (u/u_edge)**2) ## δ3 integrand
    d3 = cumulative_trapezoid(y=integrand, x=y, initial=0.)[j_edge] ## δ3
    return d3

def calc_dRC(y, u, u_tau, rho=None, j_edge=None, u_edge=None, rho_edge=None, rho_wall=None, **kwargs):
    '''
    Rotta-Clauser Δ = ∫W+dy = ∫(ρe+ue+ - ρ+u+) dy
    -----
    W  = (ρe·ue - ρ·u)/(ρe·ue) = 1 - (ρu/(ρe·ue))
    W+ = (ρe·ue - ρ·u)/(ρw·uτ) = ρe+ue+ - ρ+u+
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    for val in (u_edge, rho_edge):
        if val is not None and not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    if not isinstance(u_tau,(float,np.floating)):
        raise ValueError('u_tau must be type float')
    
    ## If ρ_edge is passed, then ρ must also be passed
    if rho_edge is not None and rho is None:
        raise ValueError('ρ_edge passed but ρ not passed')
    
    if rho is None: ## no ρ passed, set ρ=const=1 (incomp.)
        rho = np.ones_like(u)
    rho = np.asarray(rho)
    
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    if rho_edge is None:
        rho_edge = rho[j_edge]
    
    ## Get/check rho_wall
    if rho_wall is None:
        rho_wall = rho[0]
    np.testing.assert_allclose(rho_wall, rho[0], rtol=1e-4)
    
    Wplus = (rho_edge*u_edge - rho*u) / (rho_wall*u_tau) ## W+
    dRC = cumulative_trapezoid(y=Wplus, x=y, initial=0.)[j_edge] ## Δ = ∫W+dy
    return dRC

def calc_I2(y, u, u_tau, rho=None, j_edge=None, u_edge=None, rho_edge=None, rho_wall=None, **kwargs):
    '''
    Second moment of the velocity defect
    I2 = ∫ (W+)^2 d(y/Δ) = ∫ (W+)^2 dη
    W+ = (ρe·ue - ρ·u)/(ρw·uτ) = ρe+ue+ - ρ+u+
    Δ  = ∫W+dy = ∫(ρe+ue+ - ρ+u+) dy
    -----
    Monkewitz Chauhan Nagib (2007) : https://doi.org/10.1063/1.2780196
    Nagib Chauhan Monkewitz (2007) : https://doi.org/10.1098/rsta.2006.1948
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    for val in (u_edge, rho_edge):
        if val is not None and not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    if not isinstance(u_tau,(float,np.floating)):
        raise ValueError('u_tau must be type float')
    
    ## If ρ_edge is passed, then ρ must also be passed
    if rho_edge is not None and rho is None:
        raise ValueError('ρ_edge passed but ρ not passed')
    
    if rho is None: ## no ρ passed, set ρ=const=1 (incomp.)
        rho = np.ones_like(u)
    rho = np.asarray(rho)
    
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    if rho_edge is None:
        rho_edge = rho[j_edge]
    
    ## Get/check rho_wall
    if rho_wall is None:
        rho_wall = rho[0]
    np.testing.assert_allclose(rho_wall, rho[0], rtol=1e-4)
    
    ## W+
    Wplus = (rho_edge*u_edge - rho*u) / (rho_wall*u_tau)
    
    dRC = cumulative_trapezoid(y=Wplus, x=y, initial=0.)[j_edge] ## Δ = ∫W+dy
    
    ## I2 = ∫ (W+)^2 d(y/Δ) = ∫ (W+)^2 dη = (1/Δ) ∫ (W+)^2 dy
    I2_dRC = cumulative_trapezoid(y=Wplus**2, x=y, initial=0.)[j_edge]
    
    I2 = I2_dRC / dRC
    
    return I2

def calc_I3(y, u, u_tau, rho=None, j_edge=None, u_edge=None, rho_edge=None, rho_wall=None, **kwargs):
    '''
    Third moment of the velocity defect
    -----
    I3 = ∫ (W+)^3 d(y/Δ) = ∫ (W+)^3 dη
    W+ = (ρe·ue - ρ·u)/(ρw·uτ) = ρe+ue+ - ρ+u+
    Δ  = ∫W+dy = ∫(ρe+ue+ - ρ+u+) dy
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    for val in (u_edge, rho_edge):
        if val is not None and not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    if not isinstance(u_tau,(float,np.floating)):
        raise ValueError('u_tau must be type float')
    
    ## If ρ_edge is passed, then ρ must also be passed
    if rho_edge is not None and rho is None:
        raise ValueError('ρ_edge passed but ρ not passed')
    
    if rho is None: ## no ρ passed, set ρ=const=1 (incomp.)
        rho = np.ones_like(u)
    rho = np.asarray(rho)
    
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    if rho_edge is None:
        rho_edge = rho[j_edge]
    
    ## Get/check rho_wall
    if rho_wall is None:
        rho_wall = rho[0]
    np.testing.assert_allclose(rho_wall, rho[0], rtol=1e-4)
    
    ## W+
    Wplus = (rho_edge*u_edge - rho*u) / (rho_wall*u_tau)
    
    dRC = cumulative_trapezoid(y=Wplus, x=y, initial=0.)[j_edge] ## Δ = ∫W+dy
    
    ## I3 = ∫ (W+)^3 d(y/Δ) = (1/Δ) ∫ (W+)^3 dy
    I3_dRC = cumulative_trapezoid(y=Wplus**3, x=y, initial=0.)[j_edge]
    
    I3 = I3_dRC / dRC
    
    return I3

# ======================================================================

def calc_wake_parameter_1d(yplus, uplus, dplus, uplus_delta=None, **kwargs):
    '''
    Calculate the Coles wake parameter Π (1D kernel)
    
    Π = (κ/2)·[u+(δ) - (1/κ)·ln(δ+) - B]
    
    Coles profile:
      u+ = (1/κ)·ln(y+) + B + (2Π/κ)·W(y/δ)
        at y=δ, W(y/δ)==1
    
    Within the scope of this function, δ is generic i.e. not necessarily δ99
    
    Refs:
    -----
    Coles (1956)            : https://doi.org/10.1017/S0022112056000135
    Pirozzoli (2004)        : https://doi.org/10.1063/1.1637604
    Smits & Dussauge (2006) : https://doi.org/10.1007/b137383
    Chauhan et al. (2009)   : https://doi.org/10.1088/0169-5983/41/2/021404
    Nagib et al. (2007)     : https://doi.org/10.1098/rsta.2006.1948
    '''
    
    ## Von Kármán constant (κ)
    k = kwargs.get('k',0.384)
    
    ## Log-law intercept (B) : u+ = (1/κ)·ln(y+) + B
    ## see Nagib et al. (2007)
    B = kwargs.get('B',4.173)
    
    interp_kind = kwargs.get('interp_kind','cubic') ## 'linear','cubic'
    
    ## Check input y+,u+
    if (yplus.ndim!=1):
        raise ValueError('yplus must be 1D')
    if (uplus.ndim!=1):
        raise ValueError('uplus must be 1D')
    if (uplus.shape[0]!=yplus.shape[0]):
        raise ValueError("u,y shapes don't match")
    if not np.all(np.diff(yplus) > 0): ## check (+) monotonicity
        raise ValueError("y must be monotonically increasing")
    
    ## Check input δ+ and optional input u+(δ)
    if not isinstance(dplus,(float,np.floating)):
        raise ValueError('dplus must be type float')
    if uplus_delta is not None and not isinstance(uplus_delta,(float,np.floating)):
        raise ValueError('uplus_delta must be type float')
    
    ## u_{log}^+(δ) : Extrapolated u^+(δ) of log-law
    uplus_delta_loglaw = (1/k)*np.log(dplus) + B
    
    ## u+(δ)
    uplus_delta_func = interp1d(yplus, uplus, kind=interp_kind, bounds_error=True)
    uplus_delta_sol  = uplus_delta_func(dplus)
    
    ## If 'uplus_delta' was supplied, perform consistency check
    if uplus_delta is not None:
        np.testing.assert_allclose(uplus_delta_sol, uplus_delta, rtol=1e-5)
    
    uplus_delta = uplus_delta_sol
    
    Π = (k/2)*( uplus_delta - uplus_delta_loglaw )
    
    return Π

def calc_profile_edge_1d(y, ddy_u, **kwargs):
    '''
    Determine the edge index of a 1D boundary-layer profile.
    The edge is defined as the first location where |du/dy|<ϵ.
    
    Note: |-ωz|<ϵ can be found (instead of |du/dy|<ϵ) by passing -ωz as arg 'ddy_u'
    
    Note: The 'ddy_u' arg is automatically normalized by ddy_u[0], whereby ddy_u[0]==1.
      This is effectively a transformation to pseudo-'+' units. The purpose of this is
      to keep consistent sensitivity for the ϵ parameter.
    
    Parameters
    ----------
    y : np.ndarray
        1D wall-normal coordinate (units arbitrary)
    ddy_u : np.ndarray
        Precomputed du/dy
    
    Keyword Arguments
    -----------------
    epsilon : float, default=5e-5
        Edge-detection threshold for |du/dy| (or |ωz|)
    check_eps : bool, default=True
        Abort when no point satisfies |du/dy|<ϵ
    
    Returns
    -------
    y_edge : float
        Wall-normal location of the boundary-layer edge
    j_edge : int
        Index of boundary-layer edge
    '''
    
    epsilon   = kwargs.get('epsilon',5e-5)
    check_eps = kwargs.get('check_eps',True)
    
    ## Checks
    _check_validity_yx(y,ddy_u)
    n = y.shape[0]
    
    ## !!! Normalize such that |du/dy|_w == 1 !!!
    ## This is required for ϵ value/sensitivity to be consistent
    ddy_u = np.copy( ddy_u / ddy_u[0] )
    
    # === Determine j_edge
    
    ## Static ϵ
    j_edge = n-1
    for j in range(n):
        if np.abs(ddy_u[j]) < epsilon: ## |d□/dy|<ϵ
            j_edge = j
            break
    
    # ## Variable ϵ
    # keep_going = True
    # while keep_going:
    #     j_edge = n-1
    #     for j in range(n):
    #         if np.abs(ddy_u[j]) < epsilon: ## |d□/dy|<ϵ
    #             j_edge = j
    #             break
    #     if j_edge==n-1: ## Got to end, recalibrate ϵ & keep going
    #         epsilon *= 1.05
    #         msg = f'[WARNING] Recalibrating: ϵ={epsilon:0.3e}'
    #         #print(msg)
    #         tqdm.write(msg)
    #     else:
    #         keep_going = False
    
    if j_edge<3: ## Less than 3 points in profile
        raise ValueError('j_edge<3')
    if np.abs(ddy_u[j_edge])>=epsilon and check_eps:
        print('\n')
        print(f'[ERROR] j_edge={j_edge:d}')
        print(f'[ERROR] abs(ddy_u[{j_edge:d}])={np.abs(ddy_u[j_edge])}, ϵ={epsilon:0.3e}')
        print('\n')
        raise ValueError
    
    ## First point satisfying |d□/dy|<ϵ
    y_edge = float( y[j_edge] )
    
    if False: ## LEGACY: Find interpolated |d□/dy|==ϵ
        
        intrp_func = PchipInterpolator(y, ddy_u, axis=0, extrapolate=False)
        
        def __f_opt_edge_locator(y_test, intrp_func, epsilon):
            ddy_u_test = intrp_func(y_test)
            root = np.abs( ddy_u_test ) - epsilon
            return root
        
        sol = least_squares(
                    fun=__f_opt_edge_locator,
                    args=(intrp_func,epsilon,),
                    x0 = y[0] + 0.99*(y[j_edge]-y[0]),
                    xtol=1e-12,
                    ftol=1e-12,
                    gtol=1e-12,
                    method='trf',
                    bounds=(y[0], y[j_edge]),
                    )
        if not sol.success:
            raise ValueError
        if ( sol.x.shape[0] != 1 ):
            raise ValueError
        
        y_edge = float(sol.x[0])
        
        if ( y_edge > y[j_edge] ):
            raise ValueError
    
    ## Debug plot
    # import matplotlib.pyplot as plt
    # plt.close('all')
    # fig1 = plt.figure(figsize=(2*(16/9),2), dpi=300)
    # ax1 = plt.gca()
    # ax1.set_xscale('log', base=10)
    # ax1.plot(ddy_u, y, lw=0.8, )
    # #ax1.set_xlim(1e-30, 1e1)
    # ax1.axhline(y=y_edge, linestyle='dashed', c='gray', zorder=1, lw=0.5)
    # ax1.axvline(x=epsilon, linestyle='dashed', c='gray', zorder=1, lw=0.5)
    # fig1.tight_layout(pad=0.25)
    # plt.show()
    
    return y_edge, j_edge

def calc_d99_1d(y, u, j_edge=None, u_edge=None, **kwargs):
    '''
    Determine δ99 location of 1D profile
    - y : 1D coordinate vector
    - u : Profile variable (streamwise velocity u, pseudovelocity ŭ)
    - j_edge : (optional) edge index
    - u_edge : (optional) edge value
    '''
    interp_kind = kwargs.get('interp_kind','cubic')
    check = kwargs.get('check',True)
    rtol = kwargs.get('rtol',1e-4)
    
    y = np.asarray(y)
    u = np.asarray(u)
    
    ## Assert 0D scalars
    if u_edge is not None and not isinstance(u_edge, (float,np.floating)):
        raise ValueError('u_edge must be None or float')
    
    _check_validity_yx(y,u)
    n = y.shape[0]
    j_edge, u_edge = _determine_ju_edge(y,u,j_edge,u_edge,check=check,rtol=rtol)
    _check_validity_jn(j_edge,n)
    y_edge = y[j_edge]
    
    ## Index vs slice semantics
    ## If e.g. j_edge==-1, then [:j_edge] doesnt take last value, whereas
    ##   if j_edge==n, then [:j_edge] will take last value. This step makes
    ##   a 'slice' selector 'sl' which is negative-index consistent.
    if j_edge == -1:
        sl = slice(None) ## Take everything
    else:
        sl = slice(j_edge+1) ## Inclusive semantics
    
    ## Primary u(y) interpolation callable
    intrp_func = interp1d(
        x=y[sl],
        y=u[sl],
        kind=interp_kind,
        bounds_error=True,
        )
    
    ## Debug plot
    #import matplotlib as mpl
    #import matplotlib.pyplot as plt
    #plt.close('all')
    #fig1 = plt.figure(figsize=(5,5), dpi=200)
    #ax1 = fig1.gca()
    #ax1.plot(y[sl],u[sl],marker='o')
    #ax1.plot(y[sl],intrp_func(y[sl]),c='pink')
    
    ## Does the solution/intersection actually exist?
    if u[sl].max() < u_edge*0.99:
        print('\n')
        print(f'[ERROR] u.max()     = {u[sl].max():0.12f}')
        print(f'[ERROR] u_edge*0.99 = {u_edge*0.99:0.12f}')
        print('[ERROR] No solution exists: u.max() < u_edge*0.99')
        print('\n')
        raise RuntimeError
    
    ## Optimizer: callable root function for determining δ99
    def __f_opt_d99_locator(y_test, intrp_func, u_edge):
        root = 0.99*u_edge - intrp_func(y_test)
        return root
    
    # ## Perform least-squares solve
    # sol = least_squares(
    #         fun=__f_opt_d99_locator,
    #         args=(intrp_func,u_edge),
    #         x0=0.,
    #         xtol=1e-12,
    #         ftol=1e-12,
    #         gtol=1e-12,
    #         method='trf',
    #         bounds=(y.min(), y_edge),
    #         )
    # if not sol.success:
    #     raise ValueError(sol.message)
    # d99 = float(sol.x[0]) ## δ99
    
    # === Get 'root_scalar' bracket indices
    
    eps = np.finfo(u.dtype).eps * 10. ## Float comparison tolerance
    
    ## First pass: get bracket upper bound (index right above FIRST discrete crossing)
    j_bracket_high = None
    for j in range(y[sl].shape[0]):
        if u[j] - 0.99*u_edge > eps:
            j_bracket_high = j
            break
    if j_bracket_high is None:
        raise RuntimeError('Could not find bracket high')
    
    ## Second pass for lower bracket bound
    j_bracket_low = None
    for jj in range(j_bracket_high):
        j = j_bracket_high-jj
        if u[j] - 0.99*u_edge < -eps:
            j_bracket_low = j
            break
    if j_bracket_low is None:
        raise RuntimeError('Could not find bracket low')
    
    ## Perform solve with 'root_scalar'
    sol = root_scalar(
            f=__f_opt_d99_locator,
            args=(intrp_func,u_edge),
            method="brentq",
            bracket=(y[j_bracket_low],y[j_bracket_high]),
            xtol=1e-12,
            rtol=1e-12,
            )
    if not sol.converged:
        raise ValueError
    d99 = float(sol.root) ## δ99
    
    ## Check solution is within bounds
    if d99 > y_edge:
        raise ValueError('δ99 > y_edge')
    if d99 < y.min():
        raise ValueError('δ99 < y.min()')
    
    ## Debug plot
    # if not np.isclose(0.99*u_edge, u99, rtol=1e-5):
    #     ax1.axhline(y=u_edge, ls='solid', c='purple')
    #     ax1.axhline(y=0.99*u_edge, ls='solid', c='orange')
    #     ax1.axvline(x=d99, ls='solid', c='red')
    #     plt.show()
    
    ## Final assertion: u(δ99)==0.99·u_edge
    u99 = intrp_func(d99)
    np.testing.assert_allclose(0.99*u_edge, u99, rtol=1e-5)
    
    return d99

def calc_bl_integral_quantities_1d(
    y,
    u,
    rho,
    d99,
    u_tau, rho_wall, mu_wall,
    u_edge, rho_edge, mu_edge,
    j_edge,
    **kwargs,
    ):
    '''
    Convenience wrapper to calculating several integral metrics at once
    -----
    All inputs should be dimensional!
    1D vectors: y,u,rho
    0D scalars: d99, u_tau, rho_wall, mu_wall, u_edge, rho_edge, mu_edge
    0D index: j_edge
    
    Returns
    -------
    dd : dict
      δ1=δ*, δ2=θ, δ3, H12, H32, Reθ, Reτ, Reδ99, ...
    '''
    check = kwargs.get('check',True)
    rtol  = kwargs.get('rtol',1e-4)
    
    y   = np.asarray(y)
    u   = np.asarray(u)
    rho = np.asarray(rho)
    _check_validity_yx(y,u)
    _check_validity_yx(y,rho)
    n = y.shape[0]
    _check_validity_jn(j_edge,n)
    
    ## Assert 0D scalars
    for val in (d99, u_tau, rho_wall, mu_wall, u_edge, rho_edge, mu_edge):
        if not isinstance(val, (float,np.floating)):
            raise ValueError('An arg intended as 0D scalar has non-float type')
    
    # ## Consistency checks for edge quantities
    # if check:
    #     np.testing.assert_allclose(u_edge   , u[j_edge]   , rtol=rtol)
    #     np.testing.assert_allclose(rho_edge , rho[j_edge] , rtol=rtol)
    
    # === Compressible integrals (root names)
    
    d1 = calc_d1(
        y, u,
        rho=rho,
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=rho_edge,
        check=check,
        rtol=rtol,
        )
    dstar = d1
    
    d2 = calc_d2(
        y, u,
        rho=rho,
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=rho_edge,
        check=check,
        rtol=rtol,
        )
    theta = d2
    
    d3 = calc_d3(
        y, u,
        rho=rho,
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=rho_edge,
        check=check,
        rtol=rtol,
        )
    
    dRC = calc_dRC(
        y, u,
        u_tau=u_tau,
        rho=rho,
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=rho_edge,
        rho_wall=rho_wall,
        check=check,
        rtol=rtol,
        )
    
    I2 = calc_I2(
        y, u,
        u_tau=u_tau,
        rho=rho,
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=rho_edge,
        rho_wall=rho_wall,
        check=check,
        rtol=rtol,
        )
    
    I3 = calc_I3(
        y, u,
        u_tau=u_tau,
        rho=rho,
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=rho_edge,
        rho_wall=rho_wall,
        check=check,
        rtol=rtol,
        )
    
    # === Kinetic (u-only) integrals
    
    d1_k = calc_d1(
        y, u,
        rho=None, ## !! ρ ignored
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=None, ## !! ρ ignored
        check=check,
        rtol=rtol,
        )
    dstar_k = d1_k
    
    d2_k = calc_d2(
        y, u,
        rho=None, ## !! ρ ignored
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=None, ## !! ρ ignored
        check=check,
        rtol=rtol,
        )
    theta_k = d2_k
    
    d3_k = calc_d3(
        y, u,
        rho=None, ## !! ρ ignored
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=None, ## !! ρ ignored
        check=check,
        rtol=rtol,
        )
    
    dRC_k = calc_dRC(
        y, u,
        u_tau=u_tau,
        rho=None, ## !! ρ ignored
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=None, ## !! ρ ignored
        rho_wall=None, ## !! ρ ignored
        check=check,
        rtol=rtol,
        )
    
    I2_k = calc_I2(
        y, u,
        u_tau=u_tau,
        rho=None, ## !! ρ ignored
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=None, ## !! ρ ignored
        rho_wall=None, ## !! ρ ignored
        check=check,
        rtol=rtol,
        )
    
    I3_k = calc_I3(
        y, u,
        u_tau=u_tau,
        rho=None, ## !! ρ ignored
        j_edge=j_edge,
        u_edge=u_edge,
        rho_edge=None, ## !! ρ ignored
        rho_wall=None, ## !! ρ ignored
        check=check,
        rtol=rtol,
        )
    
    ## Shape factors
    H12    = d1   / d2
    H12_k  = d1_k / d2_k
    H32    = d3   / d2
    H32_k  = d3_k / d2_k
    
    ## Reynolds numbers
    Re_tau     = d99     * u_tau  * rho_wall / mu_wall ## Reτ   = δ99·uτ·ρw / μw
    Re_theta   = theta   * u_edge * rho_edge / mu_edge ## Reθ   = θ·ue·ρe / μe
    Re_dstar   = dstar   * u_edge * rho_edge / mu_edge ## Reδ*  = δ*·ue·ρe / μe
    Re_d2      = theta   * u_edge * rho_edge / mu_wall ## Reδ2  = θ·ue·ρe / μw
    Re_d99     = d99     * u_edge * rho_edge / mu_edge ## Reδ99 = δ99·ue·ρe / μe
    Re_theta_k = theta_k * u_edge * rho_edge / mu_edge ## Reθk  = θk·ue·ρe / μe
    Re_dstar_k = dstar_k * u_edge * rho_edge / mu_edge ## Reδ*k = δ*k·ue·ρe / μe
    
    ## Dictionary to return
    dd = {
        'd1':d1,
        'd1_k':d1_k,
        'dstar':dstar,
        'dstar_k':dstar_k,
        
        'd2':d2,
        'd2_k':d2_k,
        'theta':theta,
        'theta_k':theta_k,
        
        'd3':d3,
        'd3_k':d3_k,
        
        'dRC':dRC,
        'dRC_k':dRC_k,
        
        'H12':H12,
        'H12_k':H12_k,
        'H32':H32,
        'H32_k':H32_k,
        
        'I2':I2,
        'I2_k':I2_k,
        'I3':I3,
        'I3_k':I3_k,
        
        'Re_tau':Re_tau,
        'Re_theta':Re_theta,
        'Re_dstar':Re_dstar,
        'Re_d2':Re_d2,
        'Re_d99':Re_d99,
        'Re_theta_k':Re_theta_k,
        'Re_dstar_k':Re_dstar_k,
        }
    
    return dd



import numpy as np
from scipy.linalg import solve

# ======================================================================

def gradient(u, x=None, d=1, axis=0, acc=6, edge_stencil='full', return_coeffs=False, no_warn=True):
    '''
    Numerical Gradient Approximation Using Finite Differences
    -----
    - calculates stencil given arbitrary accuracy & derivative order
    - handles non-uniform grids
    - accuracy order is only mathematically valid for:
       - uniform coordinate array
       - inner points which have full central stencil
    - handles N-D numpy arrays (gradient performed over axis denoted by axis arg)
    -----
    u    : input array to perform differentiation upon
    x    : coordinate vector (np.ndarray) OR dx (float) in the case of a uniform grid
    d    : derivative order
    axis : axis along which to perform gradient
    acc  : accuracy order (only fully valid for inner points with central stencil on uniform grid)
    -----
    edge_stencil  : type of edge stencil to use ('half','full')
    return_coeffs : if True, then return stencil & coefficient information
    -----
    # stencil_npts : number of index pts in (central) stencil
    #     --> no longer an input
    #     --> using 'acc' (accuracy order) instead and calculating npts from formula
    #     - stencil_npts=3 : stencil=[      -1,0,+1      ]
    #     - stencil_npts=5 : stencil=[   -2,-1,0,+1,+2   ]
    #     - stencil_npts=7 : stencil=[-3,-2,-1,0,+1,+2,+3]
    #     - edges are filled out with appropriate clipping of central stencil
    -----
    turbx.gradient( u , x , d=1 , acc=2 , edge_stencil='half' , axis=0 )
    ...reproduces...
    np.gradient(u, x, edge_order=1, axis=0)
    
    turbx.gradient( u , x , d=1 , acc=2 , edge_stencil='full' , axis=0 )
    ...reproduces...
    np.gradient(u, x, edge_order=2, axis=0)
    -----
    https://en.wikipedia.org/wiki/Finite_difference_coefficient
    https://web.media.mit.edu/~crtaylor/calculator.html
    -----
    Fornberg B. (1988) Generation of Finite Difference Formulas on
    Arbitrarily Spaced Grids, Mathematics of Computation 51, no. 184 : 699-706.
    http://www.ams.org/journals/mcom/1988-51-184/S0025-5718-1988-0935077-0/S0025-5718-1988-0935077-0.pdf
    '''
    
    u = np.asanyarray(u)
    nd = u.ndim
    
    # print('contiguous   : %s'%str( u.data.contiguous   ) )
    # print('C-contiguous : %s'%str( u.data.c_contiguous ) )
    # print('F-contiguous : %s'%str( u.data.f_contiguous ) )
    
    if (nd==0):
        raise ValueError('turbx.gradient() requires input that is at least 1D')
    
    axes = tuple(range(nd))
    
    if not isinstance(axis, int):
        raise ValueError('axis should be of type int')
    if (axis not in axes):
        raise ValueError('axis=%i is not valid for array with u.ndim=%s'%(axis,str(u.ndim)))
    
    nx = u.shape[axis] ## size of axis over which gradient will be performed
    
    if (nx<3):
        raise ValueError('nx<3')
    
    if (x is not None):
        if isinstance(x, float):
            if (x<=0.):
                raise ValueError('if x is a float it should be >0.')
        elif isinstance(x, int):
            x = float(x)
        elif isinstance(x, np.ndarray):
            if (x.ndim!=1):
                raise ValueError('x should be 1D if it is of type np.ndarray')
            if (x.shape[0]!=nx):
                raise ValueError('size of x does not match data axis specified')
            if (not np.all(np.diff(x) > 0.)) and (not np.all(np.diff(x) < 0.)):
                    raise AssertionError('x is not monotonically increasing/decreasing')
            
            ## optimization: check if x is actually uniformly spaced, in which case x=Δx
            dx0 = x[1]-x[0]
            if np.all(np.isclose(np.diff(x), dx0, rtol=1e-12)): 
                #print('turbx.gradient() : x arr with x.shape=%s seems like it is actually uniformly spaced. applying x=%0.8e'%(str(x.shape),dx0))
                x = dx0
        
        else:
            raise ValueError('x should be a 1D np.ndarray or float')
    else:
        x = 1. ## if x not provided, assume uniform unit coordinate vector
    
    if isinstance(x, float):
        uniform_grid = True
    elif isinstance(x, np.ndarray):
        uniform_grid = False
    else:
        raise ValueError('turbx.gradient() : this should never happen... check!')
    
    if not isinstance(d, int):
        raise ValueError('d (derivative order) should be of type int')
    if not (d>0):
        raise ValueError('d (derivative order) should be >0')
    
    if not isinstance(acc, int):
        raise ValueError('acc (accuracy order) should be of type int')
    if not (acc>=2):
        raise ValueError('acc (accuracy order) should be >=2')
    if (acc%2!=0):
        raise ValueError('acc (accuracy order) should be an integer multiple of 2')
    
    ## for the d'th derivative with accuracy=acc, the following formula gives the n pts of the (central) stencil
    stencil_npts = 2*int(np.floor((d+1)/2)) - 1 + acc
    
    if not isinstance(stencil_npts, int):
        raise ValueError('stencil_npts must be of type \'int\'')
    if (stencil_npts<3):
        raise ValueError('stencil_npts should be >=3')
    if ((stencil_npts-1)%2 != 0):
        raise ValueError('(stencil_npts-1) should be divisible by 2 (for central stencil)')
    if (stencil_npts > nx):
        raise ValueError('stencil_npts > nx')
    
    if all([ (edge_stencil!='half') , (edge_stencil!='full') ]):
        raise ValueError('edge_stencil=%s not valid. options are: \'full\', \'half\''%str(edge_stencil))
    
    # ===
    
    n_full_central_stencils = nx - stencil_npts + 1
    
    if ( n_full_central_stencils < 5 ) and not no_warn:
        print('\nWARNING\n'+72*'-')
        print('n pts with full central stencils = %i (<5)'%n_full_central_stencils)
        #print('nx//3=%i'%(nx//3))
        print('--> consider reducing acc arg (accuracy order)')
        print(72*'-'+'\n')
    
    stencil_width = stencil_npts-1
    sw2           = stencil_width//2
    
    # === build up stencil & coefficients vector
    
    fdc_vec = [] ## vector of finite difference coefficient information
    
    ## left side
    for i in range(0,sw2):
        
        ## full
        stencil_L = np.arange(-i,stencil_width+1-i)
        
        ## half
        if (edge_stencil=='half') and not all([(acc==2),(d>=2)]):
            stencil_L = np.arange(-i,sw2+1)
        
        i_range = np.arange( 0 , stencil_L.shape[0] )
        
        if uniform_grid:
            fdc = fd_coeff_calculator( stencil_L , d=d , dx=x )
        else:
            fdc = fd_coeff_calculator( stencil_L , d=d , x=x[i_range] )
        
        fdc_vec.append( [ fdc , i_range , stencil_L ] )
    
    ## inner pts
    stencil = np.arange(stencil_npts) - sw2
    if uniform_grid:
        fdc_inner = fd_coeff_calculator( stencil , d=d , dx=x )
    for i in range(sw2,nx-sw2):
        
        i_range  = np.arange(i-sw2,i+sw2+1)
        
        if uniform_grid:
            fdc = fdc_inner
        else:
            fdc = fd_coeff_calculator( stencil , d=d , x=x[i_range] )
        
        fdc_vec.append( [ fdc , i_range , stencil ] )
    
    ## right side
    for i in range(nx-sw2,nx):
        
        ## full
        stencil_R = np.arange(-stencil_width+(nx-i-1),nx-i)
        
        ## half
        if (edge_stencil=='half') and not all([(acc==2),(d>=2)]):
            stencil_R = np.arange(-sw2,nx-i)
        
        i_range  = np.arange( nx-stencil_R.shape[0] , nx )
        
        if uniform_grid:
            fdc = fd_coeff_calculator( stencil_R , d=d , dx=x )
        else:
            fdc = fd_coeff_calculator( stencil_R , d=d , x=x[i_range] )
        
        fdc_vec.append( [ fdc , i_range , stencil_R ] )
    
    # === debug
    
    # print('i_range')
    # for fdc_vec_ in fdc_vec:
    #     print(fdc_vec_[1])
    # print('')
    # print('stencil')
    # for fdc_vec_ in fdc_vec:
    #     print(fdc_vec_[2])
    
    # === evaluate gradient
    
    if (nd==1): ## 1D
        
        u_ddx = np.zeros_like(u)
        for i in range(len(fdc_vec)):
            fdc, i_range, stencil = fdc_vec[i]
            u_ddx[i] = np.dot( fdc , u[i_range] )
    
    else: ## N-D
        
        ## the order in memory of the incoming data
        if u.data.contiguous:
            if u.data.c_contiguous:
                order='C'
            elif u.data.f_contiguous:
                order='F'
            else:
                raise ValueError
        else:
            order='C'
        
        ## if the array is C-ordered, use axis=0 as the axis to shift to
        ## if the array is F-ordered, use axis=-1 as the axis to shift to
        if order=='C':
            shift_pos=0 ## 0th axis
        elif order=='F':
            shift_pos=nd-1 ## last axis
        
        ## shift gradient axis
        u = np.swapaxes(u, axis, shift_pos)
        shape_new = u.shape
        if (shift_pos==0):
            size_all_but_ax = np.prod(np.array(shape_new)[1:])
        elif (shift_pos==nd-1):
            size_all_but_ax = np.prod(np.array(shape_new)[:-1])
        else:
            raise ValueError
        
        ## reshape N-D to 2D
        ## --> shift_pos=0    : gradient axis is 0, all other axes are flattened on axis=1)
        ## --> shift_pos=last : ...
        if (shift_pos==0):
            u = np.reshape(u, (nx, size_all_but_ax), order=order)
        elif (shift_pos==nd-1):
            u = np.reshape(u, (size_all_but_ax, nx), order=order)
        else:
            raise ValueError
        
        u_ddx = np.zeros_like(u,order=order)
        for i in range(nx):
            fdc, i_range, stencil = fdc_vec[i]
            ia=min(i_range)
            ib=max(i_range)+1
            if (shift_pos==0):
                u_ddx[i,:] = np.einsum('ij,i->j', u[ia:ib,:], fdc)
            else:
                u_ddx[:,i] = np.einsum('ji,i->j', u[:,ia:ib], fdc)
        
        ## reshape 2D back to original N-D
        u_ddx = np.reshape(u_ddx, shape_new, order=order)
        
        ## shift gradient axis back to original position
        u_ddx = np.swapaxes(u_ddx, shift_pos, axis)
    
    ## the original data array should have been de-referenced during this func
    u = None; del u
    
    if return_coeffs:
        return u_ddx, fdc_vec
    else:
        return u_ddx

def fd_coeff_calculator(stencil, d=1, x=None, dx=None):
    '''
    Calculate Finite Difference Coefficients for Arbitrary Stencil
    -----
    stencil : indices of stencil pts e.g. np.array([-2,-1,0,1,2])
    d       : derivative order
    x       : locations of grid points corresponding to stencil indices
    dx      : spacing of grid points in the case of uniform grid
    -----
    https://en.wikipedia.org/wiki/Finite_difference_coefficient
    https://web.media.mit.edu/~crtaylor/calculator.html
    -----
    Fornberg B. (1988) Generation of Finite Difference Formulas on
    Arbitrarily Spaced Grids, Mathematics of Computation 51, no. 184 : 699-706.
    http://www.ams.org/journals/mcom/1988-51-184/S0025-5718-1988-0935077-0/S0025-5718-1988-0935077-0.pdf
    '''
    
    stencil = np.asanyarray(stencil)
    
    if not isinstance(stencil, np.ndarray):
        raise ValueError('stencil should be of type np.ndarray')
    if (stencil.ndim!=1):
        raise ValueError('stencil should be 1D')
    if (stencil.shape[0]<2):
        raise ValueError('stencil size should be >=2')
    if (0 not in stencil):
        raise ValueError('stencil does not contain 0')
    if not np.issubdtype(stencil.dtype, np.integer):
        raise ValueError('stencil.dtype not a subdtype of np.integer')
    
    if not isinstance(d, int):
        raise ValueError('d (derivative order) should be of type int')
    if not (d>0):
        raise ValueError('d (derivative order) should be >0')
    
    if (dx is None) and (x is None):
        raise ValueError('one of args \'dx\' or \'x\' should be defined')
    if (dx is not None) and (x is not None):
        raise ValueError('only one of args \'dx\' or \'x\' should be defined')
    if (dx is not None):
        if not isinstance(dx, float):
            raise ValueError('dx should be of type float')
    
    if (x is not None):
        if not isinstance(x, np.ndarray):
            raise ValueError('x should be of type np.ndarray')
        if (x.shape[0] != stencil.shape[0]):
            raise ValueError('x, stencil should have same shape')
        if (not np.all(np.diff(x) > 0.)) and (not np.all(np.diff(x) < 0.)):
            raise AssertionError('x is not monotonically increasing/decreasing')
    
    ## overwrite stencil (int index) to be coordinate array (delta from 0 position)
    
    i0 = np.where(stencil==0)[0][0]
    
    if (x is not None):
        stencil = x - x[i0] 
    
    if (dx is not None):
        stencil = dx * stencil.astype(np.float64)
    
    nn = stencil.shape[0]
    
    dvec = np.zeros( (nn,) , dtype=np.float64 )
    #dvec = np.zeros( (nn,) , dtype=np.longdouble )
    dfac=1
    for i in range(d):
        dfac *= (i+1)
    dvec[d] = dfac
    
    ## increase precision
    #stencil = np.copy(stencil).astype(np.longdouble)
    
    #stencil_abs_max         = np.abs(stencil).max()
    stencil_abs_min_nonzero = np.abs(stencil[[ i for i in range(stencil.size) if i!=i0 ]]).min()
    
    '''
    scale/normalize the coordinate stencil (to avoid ill-conditioning)
    - if coordinates are already small/large, the Vandermonde matrix becomes
       HIGHLY ill-conditioned due to row exponents
    - coordinates are normalized here so that smallest absolute non-zero delta coord. is =1
    - RHS vector (dvec) gets normalized too
    - FD coefficients are (theoretically) unaffected
    '''
    normalize_stencil = True
    
    if normalize_stencil:
        stencil /= stencil_abs_min_nonzero
    
    mat = np.zeros( (nn,nn) , dtype=np.float64)
    #mat = np.zeros( (nn,nn) , dtype=np.longdouble)
    for i in range(nn):
        mat[i,:] = np.power( stencil , i )
    
    ## condition_number = np.linalg.cond(mat, p=-2)
    
    # mat_inv = np.linalg.inv( mat )
    # coeffv  = np.dot( mat_inv , dvec )
    
    if normalize_stencil:
        for i in range(nn):
            dvec[i] /= np.power( stencil_abs_min_nonzero , i )
    
    coeffv = solve(mat, dvec)
    
    return coeffv



import sys

import h5py
import numpy as np
from tqdm import tqdm

# ======================================================================

def h5_chunk_sizer(nxi, **kwargs):
    '''
    Solve for HDF5 dataset chunk size.

    Parameters:
    ----------
    nxi : iterable
        The shape of the full HDF5 dataset.
    constraint : iterable
        Per-axis constraint. Each element can be:
            - None         → flexible
            - int (>0)     → fixed chunk size
            - 'full' or -1 → chunk size equals axis size
            - ('max', int) → chunk size must be ≤ int
    '''
    
    size_kb    = kwargs.get('size_kb'    , 2*1024 ) ## target chunk size in [KB] --> default = 2 [MB]
    itemsize   = kwargs.get('itemsize'   , 4      ) ## dtype.itemsize --> default single precision i.e. 4 [B]
    constraint = kwargs.get('constraint' , None   ) ## iterable of nxi constraints --> int,None,'full'/-1
    base       = kwargs.get('base'       , 2      ) ## axis chunk size = ceil[size/(<int>*base)] where <int> is incremented
    
    ## if no constraint given, all axes are fully flexible
    if constraint is None:
        constraint = [ None for i in range(len(nxi)) ]
    
    ## check inputs
    if not hasattr(constraint, '__iter__') or len(nxi) != len(constraint):
        raise ValueError('nxi and constraint must be iterable and the same length')
    if not isinstance(base,int):
        raise TypeError('base must be an integer')
    if (base<1):
        raise TypeError('base must be an integer')
    
    # === increment divisor on largest axis, with divisor=<int>*base
    
    nxi = list(nxi)
    div = [ 1 for i in range(len(nxi)) ] ## divisor vector, initialize with int ones
    
    ## list of axes indices which are 'flexible' ... this is updated dynamically in loop
    i_flexible = [ i for i,c in enumerate(constraint) if c is None or isinstance(c,tuple) ]
    
    while True:
        
        div_last = list(div) ## make a copy
        #print(f'div_last = {str(tuple(div_last))}')
        
        chunks = []
        for i in range(len(nxi)):
            
            dim = nxi[i]
            
            if (constraint[i] is None):
                C = max( int(np.floor(dim/div[i])) , 1 ) ## divide by divisor
            elif (constraint[i] == 'full') or (constraint[i] == -1):
                C = dim ## chunk axis shape is == dset axis shape
            elif isinstance(constraint[i], int) and (constraint[i]>0):
                C = constraint[i] ## chunk axis shape is just the constraint
            elif isinstance(constraint[i], tuple) and (constraint[i][0]=='max') and isinstance(constraint[i][1],int):
                max_val = constraint[i][1]
                C = min( max( int(np.floor(dim/div[i])) , 1 ) , max_val )
            else:
                raise ValueError(f'problem with constraint[{i:d}] = {str(constraint[i])}')
            chunks.append(C)
        
        #print(f'chunks = {str(tuple(chunks))}')
        
        ## recalculate i_flexible
        i_flexible = []
        for i,c in enumerate(constraint):
            if chunks[i] == 1: ## already at min, is not flexible
                continue
            if c is None:
                i_flexible.append(i)
            elif isinstance(c,tuple) and c[0]=='max':
                if chunks[i] > 1:
                    i_flexible.append(i)
        
        #print(f'i_flexible = {str(i_flexible)}')
        
        ## there are no flexible axes --> exit loop
        if len(i_flexible)==0:
            break
        
        ## the current size of a chunk
        chunk_size_kb = np.prod(chunks)*itemsize / 1024.
        #print(f'chunk size {chunk_size_kb:0.1f} [KB] / {np.prod(chunks)*itemsize:d} [B]')
        
        if ( chunk_size_kb <= size_kb ): ## if chunk size is < target, then break
            break
        else: ## otherwise, increase the divisor of the greatest 'flexible' axis
            
            ## get index of (flexible) axis with greatest size
            aa = [ i for i,c in enumerate(chunks) if (i in i_flexible) ]
            bb = [ c for i,c in enumerate(chunks) if (i in i_flexible) ]
            i_gt = aa[np.argmax(bb)]
            
            ## update divisor
            div[i_gt] *= base
        
        #print(f'div = {str(tuple(div))}')
        #print('---')
        
        ## check if in infinite loop (divisor not being updated)
        if (div_last is not None) and (div == div_last):
            raise ValueError(f'invalid parameters for h5_chunk_sizer() : constraint={str(constraint)}, size_kb={size_kb:d}, base={base:d}')
    
    return tuple(chunks)

def h5_visititems_print_attrs(name, obj):
    '''
    callable for input to h5py.Group.visititems() to print names & attributes
    '''
    n_slashes = name.count('/')
    shift = n_slashes*2*' '
    item_name = name.split('/')[-1]
    
    if isinstance(obj,h5py._hl.dataset.Dataset):
        print(shift + item_name + ' --> shape=%s, dtype=%s'%( str(obj.shape), str(obj.dtype) ) )
    else:
        print(shift + item_name)
    
    ## print attributes
    for key, val in obj.attrs.items():
        try:
            print(shift + 2*' ' + f'{key} = {str(val)} --> dtype={str(val.dtype)}')
        except AttributeError:
            print(shift + 2*' ' + f'{key} = {str(val)} --> type={str(type(val).__name__)}')

def h5_print_contents(h5filehandle):
    '''
    Print file-level attributes and recursively print names & attributes of all groups and datasets.
    '''
    
    ## file-level attributes
    for key, val in h5filehandle.attrs.items():
        try:
            print(f'{key} = {str(val)} --> dtype={str(val.dtype)}')
        except AttributeError:
            print(f'{key} = {str(val)} --> type={str(type(val).__name__)}')
    
    def visitor(name, obj):
        n_slashes = name.count('/')
        shift     = n_slashes * 2 * ' '
        item_name = name.split('/')[-1]
        
        if isinstance(obj, h5py._hl.dataset.Dataset):
            print(shift + item_name + ' --> shape=%s, dtype=%s' % (str(obj.shape), str(obj.dtype)))
        else:
            print(shift + item_name)
        
        for key, val in obj.attrs.items():
            try:
                print(shift + 2 * ' ' + f'{key} = {str(val)} --> dtype={str(val.dtype)}')
            except AttributeError:
                print(shift + 2 * ' ' + f'{key} = {str(val)} --> type={str(type(val).__name__)}')
    
    # Use the visitor function with visititems
    h5filehandle.visititems(visitor)
    
    return

class h5_visit_container:
    '''
    callable for input to h5py.Group.visit() which stores dataset/group names
    '''
    def __init__(self):
        self.names = []
    def __call__(self, name):
        if (name not in self.names):
            self.names.append(name)

def h5_ds_force_allocate_chunks(ds, verbose=False, quick=False):
    '''
    Force allocation of all chunks in an ND dataset by writing real data
    '''
    if not isinstance(ds, h5py.Dataset):
        raise TypeError('ds must be a h5py.Dataset object')
    
    shape = ds.shape
    dtype = ds.dtype
    chunk_shape = ds.chunks
    rng = np.random.default_rng(seed=1)
    
    ## for contiguous datasets, fill the entire array
    if chunk_shape is None:
        #ds[...] = np.zeros(shape, dtype=dtype) ## might lead to optimizations under the hood
        #ds[...] = rng.uniform(-1,+1,size=shape).astype(dtype)
        ds[...] = rng.random(size=shape, dtype=dtype)
        return
    
    ## info needed for iterating through chunks
    chunk_starts = [range(0, dim, cdim) for dim, cdim in zip(shape, chunk_shape)]
    chunk_grid_shape = [len(r) for r in chunk_starts]
    total_chunks = np.prod(chunk_grid_shape)
    
    if verbose:
        progress_bar = tqdm(
            total=total_chunks,
            ncols=100,
            desc='allocate chunks',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    for chunk_idx in np.ndindex(*chunk_grid_shape):
        starts = [r[i] for r, i in zip(chunk_starts, chunk_idx)]
        
        if quick: ## just write a single element to allocate the chunk
            ds[tuple(starts)] = 0
        else:
            slices = tuple(
                slice(start, min(start + size, dim))
                for start, size, dim in zip(starts, chunk_shape, shape)
                )
            actual_shape = tuple(slc.stop - slc.start for slc in slices)
            #ds[slices] = np.zeros(actual_shape, dtype=dtype) ## might lead to optimizations under the hood
            #ds[slices] = rng.uniform(-1,+1,size=actual_shape).astype(dtype)
            ds[slices] = rng.random(size=actual_shape, dtype=dtype)
        
        if verbose: progress_bar.update()
    if verbose: progress_bar.close()
    return



import math
import os
import sys
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import scipy as sp
import tqdm

# ======================================================================

def get_overlapping_window_size(asz, n_win, overlap_fac):
    '''
    get window length and overlap given a
    desired number of windows and a nominal overlap factor
    -----
    --> the output should be passed to get_overlapping_windows()
        to do the actual padding & windowing
    '''
    if not isinstance(asz, (int,np.int32,np.int64)):
        raise TypeError('arg asz must be type int')
    if not isinstance(n_win, (int,np.int32,np.int64)):
        raise TypeError('arg n_win must be type int')
    if (overlap_fac >= 1.):
        raise ValueError('arg overlap_fac must be <1')
    if (overlap_fac < 0.):
        raise ValueError('arg overlap_fac must be >0')
    n_ends = n_win+1
    n_mids = n_win
    
    # === solve for float-valued window 'mid' size & 'end' size
    def eqn(soltup, asz=asz, overlap_fac=overlap_fac):
        (endsz,midsz) = soltup
        eq1 = asz - n_ends*endsz - n_mids*midsz
        eq2 = overlap_fac*(midsz+2*endsz) - endsz
        return [eq1, eq2]
    
    guess = asz*0.5
    endsz,midsz = sp.optimize.fsolve(eqn, (guess,guess), (asz,overlap_fac))
    win_len     = midsz + 2*endsz
    overlap     = endsz
    
    win_len = max(math.ceil(win_len),1)
    overlap = max(math.floor(overlap),0)
    
    return win_len, overlap

def get_overlapping_windows(a, win_len, overlap):
    '''
    subdivide 1D array into overlapping windows
    '''
    #pad_mode = kwargs.get('pad_mode','append')
    ##
    if not isinstance(a, np.ndarray):
        raise TypeError('arg a must be type np.ndarray')
    if not isinstance(win_len, int):
        raise TypeError('arg win_len must be type int')
    if not isinstance(overlap, int):
        raise TypeError('arg overlap must be type int')
    ##
    asz   = a.size
    skip  = win_len - overlap
    n_pad = (win_len - asz%skip)%skip
    #a_pad = np.concatenate(( np.zeros(n_pad,dtype=a.dtype) , np.copy(a) )) ## prepend
    a_pad = np.concatenate(( np.copy(a) , np.zeros(n_pad,dtype=a.dtype) )) ## append
    ##
    b = np.lib.stride_tricks.sliding_window_view(a_pad, win_len, axis=0)
    b = np.copy(b[::skip,:])
    n_win = b.shape[0]
    ##
    if (n_pad > 0.5*win_len):
        print('WARNING: n_pad > overlap')
    ##
    return b, n_win, n_pad

def ccor(ui,uj,**kwargs):
    '''
    1D normalized cross-correlation
    '''
    if (ui.ndim!=1):
        raise AssertionError('ui.ndim!=1')
    if (uj.ndim!=1):
        raise AssertionError('uj.ndim!=1')
    
    mode     = kwargs.get('mode','full')
    get_lags = kwargs.get('get_lags',False)
    
    if get_lags:
        lags = sp.signal.correlation_lags(ui.shape[0], uj.shape[0], mode=mode)
    else:
        lags = None
    
    #R    = sp.signal.correlate(ui, uj, mode=mode, method='direct')
    R    = sp.signal.correlate(ui, uj, mode=mode, method='fft') ## 'fft' is O(100)x faster for size O(10000) arrays
    norm = np.sqrt(np.sum(ui**2)) * np.sqrt(np.sum(uj**2))
    
    if (norm==0.):
        #R = np.ones((R.shape[0],), dtype=ui.dtype)
        R = np.zeros((R.shape[0],), dtype=ui.dtype)
    else:
        R /= norm
    
    if get_lags:
        return lags, R
    else:
        return R

def ccor_naive(u,v):
    '''
    1D normalized cross-correlation (naive version)
    - this kernel is designed as a check for ccor()
    '''
    if (u.ndim!=1):
        raise AssertionError('u.ndim!=1')
    if (v.ndim!=1):
        raise AssertionError('v.ndim!=1')
    
    ii = np.arange(u.shape[0],dtype=np.int32)
    jj = np.arange(v.shape[0],dtype=np.int32)
    
    ## lags (2D)
    ll     = np.stack(np.meshgrid(ii,jj,indexing='ij'), axis=-1)
    ll     = ll[:,:,0] - ll[:,:,1]
    
    ## lags (1D)
    lmin   = ll.min()
    lmax   = ll.max()
    n_lags = lmax-lmin+1
    lags   = np.arange(lmin,lmax+1)
    
    uu, vv = np.meshgrid(u,v,indexing='ij')
    uv     = np.stack((uu,vv), axis=-1)
    uvp    = np.prod(uv,axis=-1)
    
    c=-1
    R = np.zeros(n_lags, dtype=np.float64)
    for lag in lags:
        c+=1
        X = np.where(ll==lag)
        #N = X[0].shape[0]
        R_ = np.sum(uvp[X]) / ( np.sqrt(np.sum(u**2)) * np.sqrt(np.sum(v**2)) )
        R[c] = R_
    return lags, R

def ccor_vec(ui,uj,axis=0):
    '''
    normalized cross-correlation, vectorized wrapper
    ---
    Parameters:
        ui, uj : ndarray
            Input arrays with the same shape.
        axis : int
            Axis along which to compute the cross-correlation.
    Returns:
        R : ndarray
            Cross-correlation results normalized where norms are non-zero.
    '''
    
    if ui.shape != uj.shape:
        raise ValueError('ui and uj must have the same shape.')
    
    nd = ui.ndim
    if (nd<2):
        raise ValueError('nd<2')
    
    shape_orig = ui.shape
    axes = tuple(range(ui.ndim))
    
    if not isinstance(axis, int):
        raise ValueError('axis should be of type int')
    if (axis not in axes):
        raise ValueError(f'axis={axis:d} is not valid for array with ui.ndim={str(ui.ndim)}')
    
    ## normalization for output array
    norm = np.sqrt( np.sum( ui**2, axis=axis, keepdims=True ) ) * np.sqrt( np.sum( uj**2, axis=axis, keepdims=True ) )
    
    nx = ui.shape[axis] ## size of axis over which gradient will be performed
    shift_pos=nd-1 ## last axis
    
    ## shift cross-correlation axis to last axis
    nx = ui.shape[axis] ## size of axis over which gradient will be performed
    ui = np.swapaxes(ui, axis, shift_pos)
    uj = np.swapaxes(uj, axis, shift_pos)
    shape_new = ui.shape
    size_all_but_ax = np.prod(np.array(shape_new)[:-1])
    
    ## shape of output
    n_lags = 2*nx - 1
    shape_out = list(shape_orig)
    shape_out[axis] = n_lags
    shape_out = tuple(shape_out)
    
    ## reshape N-D to 2D
    ## cross-correlation axis is 1, all other axes are flattened on axis=0)
    ui  = np.reshape(ui, (size_all_but_ax,nx), order='C')
    uj  = np.reshape(uj, (size_all_but_ax,nx), order='C')
    
    ## vectorize kernel
    __ccor_kernel = np.vectorize(
        lambda u,v: sp.signal.correlate(u,v, mode='full', method='direct'),
        signature="(n),(n)->(m)",
    )
    
    ## run kernel
    R = __ccor_kernel(ui, uj)
    
    ## 2D to N-D
    R = np.reshape(R, (*shape_new[:-1],n_lags), order='C')
    
    ## shift cross-correlation axis back to original position
    R = np.swapaxes(R, axis, shift_pos)
    if (R.shape != shape_out):
        raise ValueError
    
    ## normalize
    mask = norm != 0
    mask = np.broadcast_to(mask, shape_out)
    norm = np.broadcast_to(norm, shape_out)
    R[mask] /= norm[mask]
    R[~mask] = 0.
    return R

def compute_bootstrap_statistic(x, f_statistic, **kwargs):
    '''
    Compute bootstrapped confidence intervals for a given statistic (function) using threads
    '''
    
    verbose          = kwargs.get('verbose',False)
    n_resamples      = kwargs.get('n_resamples',10_000)
    confidence_level = kwargs.get('confidence_level',0.99)
    max_workers      = kwargs.get('max_workers',None)
    entropy          = kwargs.get('entropy',None)
    desc             = kwargs.get('desc','compute_bootstrap_statistic()')
    
    if not isinstance(x, np.ndarray):
        raise ValueError('x must be numpy array')
    if x.ndim != 1:
        raise ValueError('x.ndim must be 1')
    
    n_samples = x.shape[0]
    sq = np.random.SeedSequence(entropy=entropy)
    seeds = sq.spawn(n_resamples) ## generate N unique child seeds for local RNGs
    # for seed_ in seeds: ## check
    #     print( int(seed_.generate_state(1)[0]) )
    def __bootstrap_worker(seed):
        local_rng = np.random.default_rng(seed=seed) ## a local RNG
        sample = local_rng.choice(x, size=n_samples, replace=True)
        return f_statistic(sample)
    
    if max_workers is None:
        max_workers = os.cpu_count()
    
    if verbose:
        progress_bar = tqdm(
            total=n_resamples,
            ncols=100,
            desc=desc,
            smoothing=0.,
            leave=False,
            file=sys.stdout,
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\n\033[F\r",
            ascii="░█",
            )
    
    stats = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = executor.map( __bootstrap_worker , seeds )
        for stat in futures:
            stats.append(stat)
            if verbose:
                progress_bar.update()
    if verbose:
        progress_bar.close()
    stats = np.array(stats, dtype=np.float64)
    alpha = (1 - confidence_level) * 100 / 2
    ci_low, ci_high = np.percentile( stats , [ alpha , 100-alpha ] )
    return ci_low, ci_high



import os
import sys
import timeit
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
import scipy as sp
from tqdm import tqdm

from .h5 import h5_print_contents
from .signal import ccor, get_overlapping_window_size, get_overlapping_windows
from .utils import even_print, format_time_string

# ======================================================================

def _calc_ccor_wall(self, **kwargs):
    '''
    Calculate cross-correlation [z,t] at every [x]
    - designed for analyzing unsteady, pre-computed wall quantities ([y] plane)
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'spd.calc_ccor_wall()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    h5py_is_mpi_build = h5py.h5.get_config().mpi
    if not h5py_is_mpi_build:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    ri = kwargs.get('ri',1)
    rj = kwargs.get('rj',1)
    rt = kwargs.get('rt',1)
    
    fn_h5_out       = kwargs.get('fn_h5_out',None)       ## filename for output HDF5 (.h5) file
    overlap_fac_nom = kwargs.get('overlap_fac_nom',0.50) ## nominal windows overlap factor
    n_win           = kwargs.get('n_win',8)              ## number of segment windows for [t] ccor calc
    
    ## check data distribution
    if (rj!=1):
        raise AssertionError('rj!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    if (ri*rj*rt != self.n_ranks):
        raise AssertionError('ri*rj*rt != self.n_ranks')
    if (ri>self.ni):
        raise AssertionError('ri>self.ni')
    if (self.ni%ri!=0):
        raise AssertionError('ni currently needs to be divisible by the n ranks')
    
    ## distribute data over [i]/[x]
    ril_ = np.array_split(np.arange(self.ni,dtype=np.int64),min(ri,self.ni))
    ril = [[b[0],b[-1]+1] for b in ril_ ]
    ri1,ri2 = ril[self.rank]
    nir = ri2 - ri1
    
    ## output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        #fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fn_h5_out_base = fname_root+'_ccor.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fn_h5_out_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'     , self.fname )
    if verbose: even_print( 'fn_h5_out' , fn_h5_out  )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    # ===
    
    ## the data dictionary to be written to .h5 later
    data = {}
    
    ## freestream data
    #lchar   = self.lchar   ; data['lchar']   = self.lchar
    U_inf   = self.U_inf   ; data['U_inf']   = self.U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = self.rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = self.T_inf
    #mu_inf  = self.mu_inf  ; data['mu_inf']  = self.mu_inf
    data['p_inf'] = self.p_inf
    data['Ma']    = self.Ma
    data['Pr']    = self.Pr
    
    ## read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    data['x'] = x
    data['z'] = z
    data['t'] = t
    
    nx = self.ni
    ni = self.ni
    nz = self.nj
    nj = self.nj
    data['ni'] = ni
    data['nx'] = nx
    data['nj'] = nj
    data['nz'] = nz
    
    nt = self.nt
    data['nt'] = nt
    
    ## assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    
    ## get Δt, dimensional [s]
    if hasattr(self,'dt'):
        dt = self.dt * self.tchar
        np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    else:
        dt = t[1] - t[0] ## already dimensional
    
    if hasattr(self,'duration'):
        t_meas = self.duration * self.tchar
        np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    else:
        t_meas = t[-1] - t[0] ## already dimensional
        self.duration = t_meas / self.tchar ## non-dimensionalize for attribute
    
    zrange = z.max() - z.min()
    
    data['t'] = t
    data['dt'] = dt
    data['dz'] = dz0
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        print(72*'-')
    
    win_len, overlap       = get_overlapping_window_size(nt, n_win, overlap_fac_nom)
    overlap_fac            = overlap / win_len
    tw, n_win, n_pad       = get_overlapping_windows(t, win_len, overlap)
    t_meas_per_win         = (win_len-1)*dt
    data['win_len']        = win_len
    data['overlap_fac']    = overlap_fac
    data['overlap']        = overlap
    data['n_win']          = n_win
    data['t_meas_per_win'] = t_meas_per_win
    
    if verbose:
        even_print('overlap_fac (nominal)' , f'{overlap_fac_nom:0.5f}'  )
        even_print('n_win'                 , f'{n_win:d}'               )
        even_print('win_len'               , f'{win_len:d}'             )
        even_print('overlap'               , f'{overlap:d}'             )
        even_print('overlap_fac'           , f'{overlap_fac:0.5f}'      )
        even_print('n_pad'                 , f'{n_pad:d}'               )
        #even_print('t_win/(δ99/uτ)'        , '%0.3f [-]'%t_eddy_per_win )
        print(72*'-')
    
    ## get lags [t]
    #lags_t,_  = ccor( np.ones(nt,dtype=np.float32), np.ones(nt,dtype=np.float32), get_lags=True )
    lags_t,_  = ccor( np.ones(win_len,dtype=np.float32), np.ones(win_len,dtype=np.float32), get_lags=True )
    #n_lags_t_ = nt*2-1
    n_lags_t_ = win_len*2-1
    n_lags_t  = lags_t.shape[0]
    if (n_lags_t!=n_lags_t_):
        raise AssertionError('check lags [t]')
    
    data['lags_t']   = lags_t
    data['n_lags_t'] = n_lags_t
    
    if verbose:
        even_print('n lags (Δt)' , '%i'%(n_lags_t,))
    
    ## get lags [z]
    lags_z,_  = ccor( np.ones(nz,dtype=np.float32) , np.ones(nz,dtype=np.float32), get_lags=True )
    n_lags_z_ = nz*2-1
    n_lags_z  = lags_z.shape[0]
    if (n_lags_z!=n_lags_z_):
        raise AssertionError('check lags [z]')
    
    data['lags_z']   = lags_z
    data['n_lags_z'] = n_lags_z
    
    if verbose:
        even_print('n lags (Δz)' , '%i'%(n_lags_z,))
    
    # ==============================================================
    # prepare buffers, etc.
    # ==============================================================
    
    do_density_weighting = False ## deactivated for now... would need implementation
    
    ## cross-correlation pairs
    ## [ str:var1, str:var2 ]
    ccor_combis = [
    
    [ 'u_tau' , 'u_tau' ], ## [ uτ , uτ ] --> ccor[ uτ′ , uτ′ ]
    [ 'v_tau' , 'v_tau' ], ## [ vτ , vτ ] --> ccor[ vτ′ , vτ′ ]
    [ 'w_tau' , 'w_tau' ], ## [ wτ , wτ ] --> ccor[ wτ′ , wτ′ ]
    
    #[ 'u_tau' , 'v_tau' ], ## [ uτ , vτ ] --> ccor[ uτ′ , vτ′ ]
    #[ 'u_tau' , 'w_tau' ], ## [ uτ , wτ ] --> ccor[ uτ′ , wτ′ ]
    #[ 'v_tau' , 'w_tau' ], ## [ vτ , wτ ] --> ccor[ vτ′ , wτ′ ]
    
    [ 'tau_uy' , 'tau_uy' ], ## [ τuy , τuy ] --> ccor[ τuy′ , τuy′ ]
    [ 'tau_vy' , 'tau_vy' ], ## [ τvy , τvy ] --> ccor[ τvy′ , τvy′ ]
    [ 'tau_wy' , 'tau_wy' ], ## [ τwy , τwy ] --> ccor[ τwy′ , τwy′ ]
    
    #[ 'tau_uy' , 'tau_vy' ], ## [ τuy , τvy ] --> ccor[ τuy′ , τvy′ ]
    #[ 'tau_uy' , 'tau_wy' ], ## [ τuy , τwy ] --> ccor[ τuy′ , τwy′ ]
    #[ 'tau_vy' , 'tau_wy' ], ## [ τvy , τwy ] --> ccor[ τvy′ , τwy′ ]
    
    [ 'p'   , 'p'   ], ## [p,p] --> ccor[ p′ , p′ ]
    [ 'T'   , 'T'   ], ## [T,T] --> ccor[ T′ , T′ ]
    [ 'rho' , 'rho' ], ## [ρ,ρ] --> ccor[ ρ′ , ρ′ ]
    
    [ 'u_tau'  , 'p' ], ## [uτ,p] --> ccor[ uτ′  , p′ ]
    [ 'tau_uy' , 'p' ], ## [τuy,p] --> ccor[ τuy′ , p′ ]
    
    ]
    
    ## generate cross-correlation scalar names
    scalars = []
    for ccor_combi in ccor_combis:
        #s1,s2,do_density_weighting = ccor_combi
        s1,s2 = ccor_combi
        if do_density_weighting:
            raise NotImplementedError
        else:
            scalars.append(f"{s1.replace('_','')}I_{s2.replace('_','')}I")
    
    scalars_dtypes = [ np.dtype(np.float64) for s in scalars ]
    
    ## generate AVG scalar names
    scalars_Re_avg = []
    #scalars_Fv_avg = []
    for ccor_combi in ccor_combis:
        #s1,s2,do_density_weighting = ccor_combi
        s1,s2 = ccor_combi
        if do_density_weighting and ('rho' not in scalars_Re_avg):
            scalars_Re_avg.append('rho')
        if do_density_weighting:
            #if (s1 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s1)
            #if (s2 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s2)
            raise NotImplementedError
        else:
            if (s1 not in scalars_Re_avg):
                scalars_Re_avg.append(s1)
            if (s2 not in scalars_Re_avg):
                scalars_Re_avg.append(s2)
    
    ## numpy formatted arrays: buffers for PSD & other data (rank-local)
    Rz         = np.zeros(shape=(nir, n_lags_z ) , dtype={'names':scalars, 'formats':scalars_dtypes})
    Rt         = np.zeros(shape=(nir, n_lags_t ) , dtype={'names':scalars, 'formats':scalars_dtypes})
    covariance = np.zeros(shape=(nir,          ) , dtype={'names':scalars, 'formats':scalars_dtypes})
    avg_Re     = np.zeros(shape=(nir,          ) , dtype={'names':scalars_Re_avg, 'formats':[ np.dtype(np.float64) for s in scalars_Re_avg ]})
    #avg_Fv     = np.zeros(shape=(nir,          ) , dtype={'names':scalars_Fv_avg, 'formats':[ np.dtype(np.float64) for s in scalars_Fv_avg ]})
    
    if verbose:
        even_print('n cross-correlation scalar combinations' , '%i'%(len(ccor_combis),))
        print(72*'-')
    
    ## window for [z] -- rectangular because [z] is assumed periodic already
    window_z       = np.ones(nz,dtype=np.float64)
    mean_sq_win_z  = np.mean(window_z**2)
    if verbose:
        even_print('mean(window_z**2)', '%0.5f'%(mean_sq_win_z,))
    
    ## window function for [t]
    window_type = None
    if (window_type=='tukey'):
        window_t = sp.signal.windows.tukey(win_len,alpha=0.5,sym=False) ## α=0:rectangular, α=1:Hann
    elif (window_type=='hann'):
        window_t = sp.signal.windows.hann(win_len,sym=False)
    elif (window_type is None):
        window_t = np.ones(win_len, dtype=np.float64)
    else:
        raise ValueError
    
    if verbose:
        even_print('window type [t]', '\'%s\''%str(window_type))
    
    ## Not needed for normalization for cross-correlation
    ##   ... but report anyway
    mean_sq_win_t = np.mean(window_t**2)
    if verbose:
        even_print('mean(window_t**2)', '%0.5f'%(mean_sq_win_t,))
    
    # ==============================================================
    # Main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            total=len(ccor_combis)*nir,
            ncols=100,
            desc='ccor',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    for cci,cc in enumerate(ccor_combis): ## ccor pairs
        
        if verbose: tqdm.write(72*'-')
        
        scalar_L, scalar_R = cc
        
        msg = f'ccor[{scalar_L}′,{scalar_R}′]'
        if verbose:
            tqdm.write(even_print('computing',msg,s=True,))
        
        dset_L = self[f'data/{scalar_L}']
        dset_R = self[f'data/{scalar_R}']
        data_gb_1x = self.n_ranks * 1 * self.nj * self.nt * dset_L.dtype.itemsize / 1024**3
        
        scalar = scalars[cci]
        
        ## assert scalar name
        scalar_ = f"{scalar_L.replace('_','')}I_{scalar_R.replace('_','')}I"
        if (scalar != scalar_):
            raise ValueError
        
        ## assert scalar name
        if do_density_weighting:
            #if (f'r{scalar_L}II_r{scalar_R}II' != scalar ):
            #    raise ValueError
            raise NotImplementedError
        else:
            if (f"{scalar_L.replace('_','')}I_{scalar_R.replace('_','')}I" != scalar ):
                raise ValueError
        
        ## [x] loop --> each rank works on part of global [x]/[i] range
        for ii in range(ri1,ri2):
            
            iii = ii - ri1 ## [x] index local
            
            data_L = np.zeros( (nz,nt) , dtype=np.float64 )
            data_R = np.zeros( (nz,nt) , dtype=np.float64 )
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## read data L
            n_scalars_read = 1 ## initialize
            scalar_str = scalar_L ## initialize
            with dset_L.collective:
                data_L[:,:] = np.copy( dset_L[ii,:,:] ).astype(np.float64)
            self.comm.Barrier()
            
            ## read data R (if != data L)
            if (scalar_L==scalar_R):
                data_R[:,:] = np.copy( data_L )
            else:
                n_scalars_read += 1
                scalar_str += f',{scalar_R}'
                with dset_R.collective:
                    data_R[:,:] = np.copy( dset_R[ii,:,:] ).astype(np.float64)
                self.comm.Barrier()
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = data_gb_1x * n_scalars_read
            
            if verbose:
                tqdm.write(even_print(f'read: {scalar_str}', f'{data_gb:0.3f} [GB]  {t_delta:0.3f} [s]  {data_gb/t_delta:0.3f} [GB/s]', s=True))
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === redimensionalize
            
            if scalar_L in ['tau_uy','tau_vy','tau_wy',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['u_tau','v_tau','w_tau',]:
                data_L *= U_inf
            elif scalar_L in ['p',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['T',]:
                data_L *= T_inf
            elif scalar_L in ['rho',]:
                data_L *= rho_inf
            else:
                raise ValueError
            
            if scalar_R in ['tau_uy','tau_vy','tau_wy',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['u_tau','v_tau','w_tau',]:
                data_R *= U_inf
            elif scalar_R in ['p',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['T',]:
                data_R *= T_inf
            elif scalar_R in ['rho',]:
                data_R *= rho_inf
            else:
                raise ValueError
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === compute mean-removed data
            
            ## avg(□) or avg(ρ·□)/avg(ρ) in [t]
            if do_density_weighting:
                #rho_avg     = np.mean(        rho , axis=-1, dtype=np.float64, keepdims=True)
                #data_L_avg  = np.mean( rho*data_L , axis=-1, dtype=np.float64, keepdims=True)
                #data_L_avg /= rho_avg
                #data_R_avg  = np.mean( rho*data_R , axis=-1, dtype=np.float64, keepdims=True)
                #data_R_avg /= rho_avg
                raise NotImplementedError
            else:
                data_L_avg = np.mean( data_L , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
                data_R_avg = np.mean( data_R , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
            
            ## data_L_avg and data_R_avg shape should be (nz,1)
            if ( data_L_avg.shape != (nz,1) ) or ( data_R_avg.shape != (nz,1) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## Reynolds prime □′ or Favre prime □″ --> shape (nz,nt)
            data_L -= data_L_avg
            data_R -= data_R_avg
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## assert stationarity / definition averaging
            ## avg(□′)==0 or avg(ρ·□″)==0
            if do_density_weighting:
                #a_ = np.mean(rho*data_L, axis=-1, dtype=np.float64, keepdims=True)
                #b_ = np.mean(rho*data_R, axis=-1, dtype=np.float64, keepdims=True)
                raise NotImplementedError
            else:
                a_ = np.mean(data_L, axis=-1, dtype=np.float64, keepdims=True) ## average in [t] --> (nz,1)
                b_ = np.mean(data_R, axis=-1, dtype=np.float64, keepdims=True)
            if not np.allclose( a_, np.zeros_like(a_), atol=1e-6 ) or not np.allclose( b_, np.zeros_like(b_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            ## covariance: <□′□′>
            if do_density_weighting:
                #covariance_ = np.mean( rho*data_L * rho*data_R , axis=-1 , dtype=np.float64, keepdims=True)
                raise NotImplementedError
            else:
                covariance_ = np.mean( data_L*data_R , axis=-1 , dtype=np.float64, keepdims=True) ## average in [t] --> (nz,1)
            
            if ( covariance_.shape != (nz,1) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## write this chunk/scalar's covariance to covariance buffer
            ## avg over [z,t] --> np.float64()
            covariance[scalar][iii] = np.mean( covariance_ , axis=(0,1) , dtype=np.float64)
            
            ## write (rank-local) 1D [x] average
            if do_density_weighting:
                #avg_Fv[scalar_L][iii] = np.mean( data_L_avg , axis=(0,1) , dtype=np.float64) )
                #avg_Fv[scalar_R][iii] = np.mean( data_R_avg , axis=(0,1) , dtype=np.float64) )
                #avg_Re['rho'][iii]    = np.mean( rho_avg    , axis=(0,1) , dtype=np.float64) )
                raise ValueError
            else:
                avg_Re[scalar_L][iii] = np.mean( data_L_avg , axis=(0,1) , dtype=np.float64)
                avg_Re[scalar_R][iii] = np.mean( data_R_avg , axis=(0,1) , dtype=np.float64)
            
            # ===============================================================================
            # At this point you have 4D [x,y,z,t] □′ data
            # ===============================================================================
            
            
            ## do [z] cross-correlation for every [t]
            Rz_buf = np.zeros((nt,n_lags_z), dtype=np.float64)
            for ti in range(nt):
                uL = np.copy( data_L[:,ti] )
                uR = np.copy( data_R[:,ti] )
                Rz_buf[ti,:] = ccor(uL,uR)
            
            Rz[scalar][iii,:] = np.mean(Rz_buf, axis=0, dtype=np.float64) ## mean across [t] --> (n_lags_z,)
            
            
            ## ## do [t] cross-correlation for every [z]
            ## Rt_buf = np.zeros((nz,n_lags_t), dtype=np.float64)
            ## for zi in range(nz):
            ##     uL = np.copy( data_L[zi,:] )
            ##     uR = np.copy( data_R[zi,:] )
            ##     Rt_buf[zi,:] = ccor(uL,uR)
            ## 
            ## Rt[scalar][iii,:] = np.mean(Rt_buf, axis=0, dtype=np.float64) ## mean across [z] --> (n_lags_t,)
            
            
            ## do [t] cross-correlation for every [z]
            Rt_buf = np.zeros((nz,n_lags_t), dtype=np.float64)
            for zi in range(nz):
                uL = np.copy( data_L[zi,:] )
                uR = np.copy( data_R[zi,:] )
                #Rt_buf[zi,:] = ccor(uL,uR)
                
                uL_, nw, n_pad = get_overlapping_windows(uL, win_len, overlap)
                uR_, nw, n_pad = get_overlapping_windows(uR, win_len, overlap)
                
                ## Per-segment buffer
                Rt_buf_win = np.zeros((nw,n_lags_t), dtype=np.float64)
                for wi in range(nw):
                    A1 = np.copy( uL_[wi,:] * window_t ) 
                    A2 = np.copy( uR_[wi,:] * window_t )
                    Rt_buf_win[wi,:] = ccor(A1,A2)
                Rt_buf[zi,:] = np.mean(Rt_buf_win, axis=0, dtype=np.float64) ## mean across windows --> (n_lags_t,)
            
            Rt[scalar][iii,:] = np.mean(Rt_buf, axis=0, dtype=np.float64) ## mean across [z] --> (n_lags_t,)
            
            
            self.comm.Barrier() ## [x] loop ('ii' within this rank's range)
            if verbose: progress_bar.update()
            
            #break ## DEBUG
        #break ## DEBUG
    
    self.comm.Barrier()
    if verbose:
        progress_bar.close()
        print(72*'-')
    
    # ==============================================================
    # Write HDF5 (.h5) file
    # ==============================================================
    
    ## Overwrite outfile!
    ## - first, open on rank 0 and to write attributes and initialize datasets
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## write numpy arrays
            hfw.create_dataset( 'dims/x'      , data=x      ) ## [m]
            hfw.create_dataset( 'dims/z'      , data=z      ) ## [m]
            hfw.create_dataset( 'dims/t'      , data=t      ) ## [s]
            hfw.create_dataset( 'dims/lags_z' , data=lags_z )
            hfw.create_dataset( 'dims/lags_t' , data=lags_t )
            
            ## initialize datasets : covariance,Rz,Rt
            for scalar in scalars:
                hfw.create_dataset( f'covariance/{scalar}' , shape=(nx,)          , dtype=np.float64 , chunks=None         , data=np.full((nx,),0.,np.float64)         )
                hfw.create_dataset( f'Rz/{scalar}'         , shape=(nx,n_lags_z)  , dtype=np.float64 , chunks=(1,n_lags_z) , data=np.full((nx,n_lags_z),0.,np.float64) )
                hfw.create_dataset( f'Rt/{scalar}'         , shape=(nx,n_lags_t)  , dtype=np.float64 , chunks=(1,n_lags_t) , data=np.full((nx,n_lags_t),0.,np.float64) )
            
            ## initialize datasets : 1D [x] mean
            for scalar in avg_Re.dtype.names:
                hfw.create_dataset( f'avg/Re/{scalar}', shape=(nx,), dtype=np.float64, chunks=None, data=np.full((nx,),0.,np.float64) )
    
    self.comm.Barrier()
    
    ## re-open in parallel for data writes
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## collectively write covariance,Rz,Rt
        for scalar in scalars:
            dset = hfw[f'covariance/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = covariance[scalar][:]
            dset = hfw[f'Rz/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = Rz[scalar][:,:]
            dset = hfw[f'Rt/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = Rt[scalar][:,:]
        
        ## collectively write 1D [z] avgs
        for scalar in avg_Re.dtype.names:
            dset = hfw[f'avg/Re/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = avg_Re[scalar][:]
    
    ## report file contents
    self.comm.Barrier()
    if (self.rank==0):
        even_print( os.path.basename(fn_h5_out) , f'{(os.path.getsize(fn_h5_out)/1024**2):0.1f} [MB]' )
        print(72*'-')
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : spd.calc_ccor_wall() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import os
import sys
import timeit
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
import psutil
from mpi4py import MPI
from scipy.fft import fft, fftfreq, fftshift
from scipy.integrate import trapezoid
from scipy.signal.windows import hann
from tqdm import tqdm

from .h5 import h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_turb_cospectrum_wall(self, **kwargs):
    '''
    Calculate FFT cospectrum in [z,t] at every [x]
    - designed for analyzing unsteady, pre-computed wall quantities ([y] plane)
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'spd.calc_turb_cospectrum_wall()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    if not h5py.h5.get_config().mpi:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    ri = kwargs.get('ri',1)
    rj = kwargs.get('rj',1)
    rt = kwargs.get('rt',1)
    
    n_threads = kwargs.get('n_threads',1)    ## Number of threads for fft() call
    fn_h5_out = kwargs.get('fn_h5_out',None) ## Filename for output HDF5 (.h5) file (default chosen later)
    n_win     = kwargs.get('n_win',16)       ## Number of segment windows for [t] PSD calc
    
    ## Check data distribution (only distribute data in [i]/[x])
    if (rj!=1):
        raise AssertionError('rj!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    
    if not isinstance(ri,int) or (ri<1):
        raise ValueError('ri should be a positive non-zero int')
    
    if (ri*rj*rt != self.n_ranks):
        raise AssertionError('ri*rj*rt != self.n_ranks')
    if (ri>self.ni):
        raise AssertionError('ri>self.ni')
    if (self.ni%ri!=0):
        raise AssertionError('ni currently needs to be divisible by the n ranks')
    
    ## Distribute data over [i]/[x]
    ril_ = np.array_split(np.arange(self.ni,dtype=np.int64),min(ri,self.ni))
    ril = [[b[0],b[-1]+1] for b in ril_ ]
    ri1,ri2 = ril[self.rank]
    nir = ri2 - ri1
    
    ## Output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        #fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fn_h5_out_base = fname_root+'_fft.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fn_h5_out_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'     , self.fname )
    if verbose: even_print( 'fn_h5_out' , fn_h5_out  )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    # ===
    
    ## Data dictionary to be written to .h5 later
    data = {}
    
    ## Infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('ni',f'{self.ni:d}')
    if verbose: even_print('nj',f'{self.nj:d}')
    if verbose: even_print('nt',f'{self.nt:d}')
    if verbose: even_print('n_pts',f'{self.n_pts/1e6:0.1f} [M]')
    if verbose: even_print('n_ranks',f'{self.n_ranks:d}')
    if verbose: print(72*'-')
    
    ## 0D freestream scalars
    #lchar   = self.lchar   ; data['lchar']   = self.lchar
    U_inf   = self.U_inf   ; data['U_inf']   = self.U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = self.rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = self.T_inf
    #mu_inf  = self.mu_inf  ; data['mu_inf']  = self.mu_inf
    data['p_inf'] = self.p_inf
    #data['M_inf'] = self.M_inf
    data['Ma']    = self.Ma
    data['Pr']    = self.Pr
    
    ## Read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    data['x'] = x
    data['z'] = z
    data['t'] = t
    
    nx = self.ni
    ni = self.ni
    nz = self.nj
    nj = self.nj
    data['ni'] = ni
    data['nx'] = nx
    data['nj'] = nj
    data['nz'] = nz
    
    nt = self.nt
    data['nt'] = nt
    
    ## Assert constant Δz
    dz0 = float(np.diff(z)[0])
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    dz = float(np.diff(z)[0])
    
    ## Get Δt, dimensional [s]
    if hasattr(self,'dt'):
        dt = self.dt * self.tchar
        np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    else:
        dt = t[1] - t[0] ## already dimensional
    
    ## Get duration of [t] signal (dimensional)
    if hasattr(self,'duration'):
        t_meas = self.duration * self.tchar
        np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    else:
        t_meas = t[-1] - t[0] ## already dimensional
        self.duration = t_meas / self.tchar ## non-dimensionalize for attribute
    
    zrange = z.max() - z.min()
    
    data['t']      = t
    data['dt']     = dt
    data['dz']     = dz
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## Report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        print(72*'-')
    
    ## Establish [t] windowing
    ## - 50% overlap is hardcoded
    nperseg  = int(2*nt/(n_win+1))          ## nt window for 50% overlap (nominal)
    nperseg  = (nperseg//2)*2               ## make divisible by 2
    noverlap = nperseg//2                   ## nt overlap
    ntu      = nperseg + (n_win-1)*noverlap ## nt used
    overlap_fac = noverlap / nperseg
    
    ## Window for [t] FFT : Hann
    win  = hann(nperseg, sym=False).astype(np.float64) ## Hann window
    win2 = np.mean(win*win)                            ## Window power correction
    
    ## Temporal [t] frequency (f) vector --> Using windows (Welch-type)
    freq_full = fftshift(fftfreq(n=nperseg, d=dt))
    fp        = np.where(freq_full>0)[0] ## Indices of positive values
    freq      = np.copy(freq_full[fp])
    df        = float(freq[1] - freq[0])
    nf        = int(freq.shape[0])
    
    data['n_win']    = n_win
    data['nperseg']  = nperseg
    data['noverlap'] = noverlap
    data['freq']     = freq
    data['df']       = df
    data['nf']       = nf
    
    if verbose:
        even_print('n_win'          , f'{n_win:d}'          )
        even_print('nperseg'        , f'{nperseg:d}'        )
        even_print('noverlap'       , f'{noverlap:d}'       )
        even_print('overlap_fac'    , f'{overlap_fac:0.5f}' )
        even_print('timesteps used' , f'{ntu}/{nt}'         )
        print(72*'-')
    
    if verbose:
        even_print('freq min',f'{freq.min():0.1f} [Hz]')
        even_print('freq max',f'{freq.max():0.1f} [Hz]')
        even_print('df',f'{df:0.1f} [Hz]')
        even_print('nf',f'{nf:d}')
        print(72*'-')
    
    ## Spatial [z] wavenumber (kz) vector
    kz_full = fftshift(fftfreq(n=nz, d=dz)) * ( 2 * np.pi )
    kzp     = np.where(kz_full>0)[0] ## Indices of positive values
    kz      = np.copy(kz_full[kzp])
    dkz     = float(kz[1] - kz[0])
    nkz     = int(kz.shape[0])
    
    data['kz']  = kz
    data['dkz'] = dkz
    data['nkz'] = nkz
    
    ## λz = 2·π/kz
    lz  = np.copy( 2 * np.pi / kz ) ## Order is descending, but matches Pkz output
    dlz = abs(float(lz[1] - lz[0]))
    nlz = int(lz.shape[0])
    
    data['lz']  = lz
    data['dlz'] = dlz
    data['nlz'] = nlz
    
    if verbose:
        even_print('kz min',f'{kz.min():0.1f} [1/m]')
        even_print('kz max',f'{kz.max():0.1f} [1/m]')
        even_print('dkz',f'{dkz:0.1f} [1/m]')
        even_print('nkz',f'{nkz:d}')
        print(72*'-')
    
    # ==================================================================
    
    do_density_weighting = False ## deactivated for now... would need implementation
    
    ## Cospectrum pairs
    ## [ str:var1, str:var2 ]
    fft_combis = [
    
    [ 'u_tau' , 'u_tau' ], ## [ uτ , uτ ] --> FFT[ uτ′ , uτ′ ]
    [ 'v_tau' , 'v_tau' ], ## [ vτ , vτ ] --> FFT[ vτ′ , vτ′ ]
    [ 'w_tau' , 'w_tau' ], ## [ wτ , wτ ] --> FFT[ wτ′ , wτ′ ]
    
    #[ 'u_tau' , 'v_tau' ], ## [ uτ , vτ ] --> FFT[ uτ′ , vτ′ ]
    #[ 'u_tau' , 'w_tau' ], ## [ uτ , wτ ] --> FFT[ uτ′ , wτ′ ]
    #[ 'v_tau' , 'w_tau' ], ## [ vτ , wτ ] --> FFT[ vτ′ , wτ′ ]
    
    [ 'tau_uy' , 'tau_uy' ], ## [ τuy , τuy ] --> FFT[ τuy′ , τuy′ ]
    [ 'tau_vy' , 'tau_vy' ], ## [ τvy , τvy ] --> FFT[ τvy′ , τvy′ ]
    [ 'tau_wy' , 'tau_wy' ], ## [ τwy , τwy ] --> FFT[ τwy′ , τwy′ ]
    
    #[ 'tau_uy' , 'tau_vy' ], ## [ τuy , τvy ] --> FFT[ τuy′ , τvy′ ]
    #[ 'tau_uy' , 'tau_wy' ], ## [ τuy , τwy ] --> FFT[ τuy′ , τwy′ ]
    #[ 'tau_vy' , 'tau_wy' ], ## [ τvy , τwy ] --> FFT[ τvy′ , τwy′ ]
    
    [ 'p'   , 'p'   ], ## [p,p] --> FFT[ p′ , p′ ]
    [ 'T'   , 'T'   ], ## [T,T] --> FFT[ T′ , T′ ]
    [ 'rho' , 'rho' ], ## [ρ,ρ] --> FFT[ ρ′ , ρ′ ]
    
    [ 'u_tau'  , 'p' ], ## [uτ,p] --> FFT[ uτ′  , p′ ]
    [ 'tau_uy' , 'p' ], ## [τuy,p] --> FFT[ τuy′ , p′ ]
    
    ]
    
    ## Generate FFT cospectrum scalar names
    scalars = []
    for fft_combi in fft_combis:
        #s1,s2,do_density_weighting = fft_combi
        s1,s2 = fft_combi
        if do_density_weighting:
            raise NotImplementedError
        else:
            scalars.append(f"{s1.replace('_','')}I_{s2.replace('_','')}I")
    
    ## Generate avg scalar names
    scalars_Re_avg = []
    #scalars_Fv_avg = []
    for fft_combi in fft_combis:
        #s1,s2,do_density_weighting = fft_combi
        s1,s2 = fft_combi
        if do_density_weighting and ('rho' not in scalars_Re_avg):
            scalars_Re_avg.append('rho')
        if do_density_weighting:
            #if (s1 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s1)
            #if (s2 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s2)
            raise NotImplementedError
        else:
            if (s1 not in scalars_Re_avg):
                scalars_Re_avg.append(s1)
            if (s2 not in scalars_Re_avg):
                scalars_Re_avg.append(s2)
    
    ## numpy formatted arrays: buffers for PSD & other data (rank-local)
    Pkz        = np.zeros(shape=(nir,nkz ) , dtype={'names':scalars        , 'formats':[ np.dtype(np.float64) for s in scalars        ]})
    Pf         = np.zeros(shape=(nir,nf  ) , dtype={'names':scalars        , 'formats':[ np.dtype(np.float64) for s in scalars        ]})
    covariance = np.zeros(shape=(nir,    ) , dtype={'names':scalars        , 'formats':[ np.dtype(np.float64) for s in scalars        ]})
    avg_Re     = np.zeros(shape=(nir,    ) , dtype={'names':scalars_Re_avg , 'formats':[ np.dtype(np.float64) for s in scalars_Re_avg ]})
    #avg_Fv     = np.zeros(shape=(nir,    ) , dtype={'names':scalars_Fv_avg , 'formats':[ np.dtype(np.float64) for s in scalars_Fv_avg ]})
    
    if verbose:
        even_print('n turb spectrum scalar combinations' , '%i'%(len(fft_combis),))
        print(72*'-')
    
    # ==================================================================
    # Check memory
    # ==================================================================
    
    hostname = MPI.Get_processor_name()
    mem_free_gb = psutil.virtual_memory().free / 1024**3
    G = self.comm.gather([ self.rank , hostname , mem_free_gb ], root=0)
    G = self.comm.bcast(G, root=0)
    
    host_mem = {}
    for rank, host, mem in G:
        if host not in host_mem or mem < host_mem[host]:
            host_mem[host] = mem
    total_free = sum(host_mem.values())
    
    if verbose:
        #print(72*'-')
        for key,value in host_mem.items():
            even_print(f'RAM free {key}', f'{int(np.floor(value)):d} [GB]')
        even_print('RAM free (local,min)', f'{int(np.floor(min(host_mem.values()))):d} [GB]')
        even_print('RAM free (global)', f'{int(np.floor(total_free)):d} [GB]')
    
    shape_read = (1,1,nz,nt) ## local
    if verbose: even_print('read shape (local)', f'[{1:d},{1:d},{nz:d},{nt:d}]')
    data_gb = np.dtype(np.float64).itemsize * np.prod(shape_read) / 1024**3
    if verbose: even_print('read size (global)', f'{int(np.ceil(data_gb*ri)):d} [GB]')
    
    if verbose: even_print('read size (global) ×8', f'{int(np.ceil(data_gb*ri*8)):d} [GB]')
    ram_usage_est = data_gb*ri*8/total_free
    if verbose: even_print('RAM usage estimate', f'{100*ram_usage_est:0.1f} [%]')
    
    self.comm.Barrier()
    if (ram_usage_est>0.80):
        print('RAM consumption might be too high. exiting.')
        self.comm.Abort(1)
    
    # ==================================================================
    # Main loop
    # ==================================================================
    
    if verbose:
        progress_bar = tqdm(
            total=len(fft_combis)*nir,
            ncols=100,
            desc='fft',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#00ffff',
            )
    
    for cci,cc in enumerate(fft_combis): ## FFT pairs
        
        if verbose: tqdm.write(72*'-')
        
        scalar_L, scalar_R = cc
        
        msg = f'FFT[{scalar_L}′,{scalar_R}′]'
        if verbose:
            tqdm.write(even_print('computing',msg,s=True,))
        
        dset_L = self[f'data/{scalar_L}']
        dset_R = self[f'data/{scalar_R}']
        data_gb_1x = self.n_ranks * 1 * self.nj * self.nt * dset_L.dtype.itemsize / 1024**3
        
        scalar = scalars[cci]
        
        ## Assert scalar name
        scalar_ = f"{scalar_L.replace('_','')}I_{scalar_R.replace('_','')}I"
        if (scalar != scalar_):
            raise ValueError
        
        ## [x] loop (rank-local)
        for ii in range(ri1,ri2):
            
            iii = ii - ri1
            
            data_L = np.zeros( (nz,nt) , dtype=np.float64 )
            data_R = np.zeros( (nz,nt) , dtype=np.float64 )
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## Read data L
            n_scalars_read = 1 ## initialize
            scalar_str = scalar_L ## initialize
            with dset_L.collective:
                data_L[:,:] = np.copy( dset_L[ii,:,:] ).astype(np.float64)
            self.comm.Barrier()
            
            ## Read data R (if != data L)
            if (scalar_L==scalar_R):
                data_R[:,:] = np.copy( data_L )
            else:
                n_scalars_read += 1
                scalar_str += f',{scalar_R}'
                with dset_R.collective:
                    data_R[:,:] = np.copy( dset_R[ii,:,:] ).astype(np.float64)
                self.comm.Barrier()
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = data_gb_1x * n_scalars_read
            
            if verbose:
                tqdm.write(even_print(f'read: {scalar_str}', f'{data_gb:0.3f} [GB]  {t_delta:0.3f} [s]  {data_gb/t_delta:0.3f} [GB/s]', s=True))
            
            ## data_L and data_R shape should be (nz,nt)
            if data_L.shape != (nz,nt) or data_R.shape != (nz,nt):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === Redimensionalize unsteady data
            
            if scalar_L in ['tau_uy','tau_vy','tau_wy',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['u_tau','v_tau','w_tau',]:
                data_L *= U_inf
            elif scalar_L in ['p',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['T',]:
                data_L *= T_inf
            elif scalar_L in ['rho',]:
                data_L *= rho_inf
            else:
                raise ValueError
            
            if scalar_R in ['tau_uy','tau_vy','tau_wy',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['u_tau','v_tau','w_tau',]:
                data_R *= U_inf
            elif scalar_R in ['p',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['T',]:
                data_R *= T_inf
            elif scalar_R in ['rho',]:
                data_R *= rho_inf
            else:
                raise ValueError
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === Compute mean-removed data
            
            ## Reynolds avg(□) or Favre avg(ρ·□)/avg(ρ) in [t]
            if do_density_weighting:
                #rho_avg     = np.mean(        rho , axis=-1, dtype=np.float64, keepdims=True)
                #data_L_avg  = np.mean( rho*data_L , axis=-1, dtype=np.float64, keepdims=True)
                #data_L_avg /= rho_avg
                #data_R_avg  = np.mean( rho*data_R , axis=-1, dtype=np.float64, keepdims=True)
                #data_R_avg /= rho_avg
                raise NotImplementedError
            else:
                data_L_avg = np.mean( data_L , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
                data_R_avg = np.mean( data_R , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
            
            ## Shape data_L_avg and data_R_avg should be (nz,1)
            if ( data_L_avg.shape != (nz,1) ) or ( data_R_avg.shape != (nz,1) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## Reynolds prime □′ or Favre prime □″ --> shape (nz,nt)
            data_L -= data_L_avg
            data_R -= data_R_avg
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## Assert stationarity + definition Re/Fv averaging
            ## avg(□′)==0 or avg(ρ·□″)==0
            if do_density_weighting:
                #a_ = np.mean(rho*data_L, axis=-1, dtype=np.float64, keepdims=True)
                #b_ = np.mean(rho*data_R, axis=-1, dtype=np.float64, keepdims=True)
                raise NotImplementedError
            else:
                a_ = np.mean(data_L, axis=-1, dtype=np.float64, keepdims=True) ## average in [t] --> (nz,1)
                b_ = np.mean(data_R, axis=-1, dtype=np.float64, keepdims=True)
            if not np.allclose( a_, np.zeros_like(a_), atol=1e-6 ) or not np.allclose( b_, np.zeros_like(b_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            ## Covariance: <□′·□′> OR <ρ□″·ρ□″> --> note that this is NOT the typical Favre <ρ·□″□″>
            if do_density_weighting:
                #covariance_ = np.mean( rho*data_L * rho*data_R , axis=-1 , dtype=np.float64, keepdims=True)
                raise NotImplementedError
            else:
                covariance_ = np.mean( data_L*data_R , axis=-1 , dtype=np.float64, keepdims=True) ## average in [t] --> (nz,1)
            
            if ( covariance_.shape != (nz,1) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## Write this chunk/scalar's covariance to covariance buffer
            ## avg over [z,t] --> np.float64()
            covariance[scalar][iii] = np.mean( covariance_ , axis=(0,1) , dtype=np.float64)
            
            ## Write (rank-local) 1D [x] average
            if do_density_weighting:
                #avg_Fv[scalar_L][iii] = np.mean( data_L_avg , axis=(0,1) , dtype=np.float64) )
                #avg_Fv[scalar_R][iii] = np.mean( data_R_avg , axis=(0,1) , dtype=np.float64) )
                #avg_Re['rho'][iii]    = np.mean( rho_avg    , axis=(0,1) , dtype=np.float64) )
                raise ValueError
            else:
                avg_Re[scalar_L][iii] = np.mean( data_L_avg , axis=(0,1) , dtype=np.float64)
                avg_Re[scalar_R][iii] = np.mean( data_R_avg , axis=(0,1) , dtype=np.float64)
            
            # ==========================================================
            # At this point you have 4D [x,y,z,t] □′ data
            # ==========================================================
            
            ## Dummy for now
            do_density_weighting_L = False
            do_density_weighting_R = False
            
            # ==========================================================
            # FFT in [t]
            # ==========================================================
            
            ## Accumulator for multiple time windows
            P = np.full((nperseg,), 0., dtype=np.float64)
            
            ## [t] Window loop
            for wi in range(n_win):
                tiA = wi*noverlap
                tiB = tiA + nperseg
                if tiB > nt:
                    raise ValueError
                
                ## [z,t]
                uL = np.full((nz,nperseg), 0., dtype=np.float64)
                uR = np.full((nz,nperseg), 0., dtype=np.float64)
                
                ## 3D (x,z,t) □′ or ρ·□″ arrays
                if do_density_weighting_L:
                    #uL[:,:] = rho[:,tiA:tiB] * data_L[:,tiA:tiB]
                    raise NotImplementedError
                else:
                    uL[:,:] = data_L[:,tiA:tiB]
                
                if do_density_weighting_R:
                    #uR[:,:] = rho[:,tiA:tiB] * data_R[:,tiA:tiB]
                    raise NotImplementedError
                else:
                    uR[:,:] = data_R[:,tiA:tiB]
                
                ## Apply Hann window to [t] dimension (for all [z])
                uL *= win[np.newaxis,:]
                uR *= win[np.newaxis,:]
                
                Ai   = fftshift( fft(uL, axis=1, workers=n_threads), axes=1) ## [t] FFT for all [z]
                Aj   = fftshift( fft(uR, axis=1, workers=n_threads), axes=1) ## [t] FFT for all [z]
                Pij  = np.real( Ai * Aj.conj() )
                Pseg = np.mean(Pij, axis=(0,)) ## [z] average
                
                ## Convert to density
                Pseg /= nperseg**2 ## nperseg is nt per window
                Pseg /= df
                
                ## Compensate window power loss
                Pseg /= win2
                
                P[:] += (1/n_win)*Pseg ## Add this window's contribution to accumulator
            
            ## Take integral over frequency
            isort    = np.argsort(freq_full)
            integral = trapezoid(P[isort], x=freq_full[isort], axis=0)
            covar    = float(covariance[scalar][iii])
            
            if False: ## DEBUG for integral / covariance regularization
                rtol = np.abs(integral-covar)/max(np.abs(max(integral,covar)),1e-12)
                if not np.isclose(covar, integral, rtol=0.005, atol=1e-5):
                    print(f'i={ii:d} FAIL : rtol={rtol:0.3e} covar={covar:0.3e} integral={integral:0.3e}')
                else:
                    print(f'i={ii:d} PASS : rtol={rtol:0.3e} covar={covar:0.3e} integral={integral:0.3e}')
            
            ## Integral / covariance regularization
            ## Scale by tiny factor to reinstate Parseval
            if np.abs(integral)>0.:
                fac = covar / integral
                P *= fac
            
            ## Write only freq>0 (for plotting etc) to output buffer
            ## Integrating the output spectrum will yield covar/2
            Pf[scalar][iii,:] = P[fp]
            
            # ==========================================================
            # FFT in [z]
            # ==========================================================
            
            ## [z,t]
            uL = np.full((nz,nt), 0., dtype=np.float64)
            uR = np.full((nz,nt), 0., dtype=np.float64)
            
            ## 3D (x,z,t) □′ or ρ·□″ arrays
            if do_density_weighting_L:
                #uL[:,:] = rho[:,:] * data_L[:,:]
                raise NotImplementedError
            else:
                uL[:,:] = data_L[:,:]
            
            if do_density_weighting_R:
                #uR[:,:] = rho[:,:] * data_R[:,:]
                raise NotImplementedError
            else:
                uR[:,:] = data_R[:,:]
            
            Ai  = fftshift( fft(uL, axis=0, workers=n_threads), axes=0) ## [z] FFT for all [t]
            Aj  = fftshift( fft(uR, axis=0, workers=n_threads), axes=0) ## [z] FFT for all [t]
            Pij = np.real( Ai * Aj.conj() )
            P   = np.mean(Pij, axis=(1,)) ## [x,t] average
            
            ## Convert to density
            P /= nz**2
            P /= dkz
            
            ## Take integral over kz
            isort    = np.argsort(kz_full)
            integral = trapezoid(P[isort], x=kz_full[isort], axis=0)
            covar    = float(covariance[scalar][iii])
            
            if False: ## DEBUG for integral / covariance regularization
                rtol = np.abs(integral-covar)/max(np.abs(max(integral,covar)),1e-12)
                if not np.isclose(covar, integral, rtol=0.005, atol=1e-5):
                    print(f'i={ii:d} FAIL : rtol={rtol:0.3e} covar={covar:0.3e} integral={integral:0.3e}')
                else:
                    print(f'i={ii:d} PASS : rtol={rtol:0.3e} covar={covar:0.3e} integral={integral:0.3e}')
            
            ## Integral / covariance regularization
            ## Scale by tiny factor to reinstate Parseval
            if np.abs(integral)>0.:
                fac = covar / integral
                P *= fac
            
            ## Write only kz>0 (for plotting etc) to output buffer
            ## Integrating the output spectrum will yield covar/2
            Pkz[scalar][iii,:] = P[kzp]
            
            # ==========================================================
            
            self.comm.Barrier() ## [x] loop ('ii' within this rank's range)
            if verbose: progress_bar.update()
            
            #break ## DEBUG
    
    self.comm.Barrier()
    if verbose:
        progress_bar.close()
        print(72*'-')
    
    # ==================================================================
    # Write HDF5 (.h5) file
    # ==================================================================
    
    ## Overwrite outfile!
    ## Open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## Write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## Write numpy arrays
            hfw.create_dataset( 'dims/x'    , data=x    ) ## [m]
            hfw.create_dataset( 'dims/z'    , data=z    ) ## [m]
            hfw.create_dataset( 'dims/t'    , data=t    ) ## [s]
            hfw.create_dataset( 'dims/freq' , data=freq ) ## [1/s] | [Hz]
            hfw.create_dataset( 'dims/kz'   , data=kz   ) ## [1/m]
            hfw.create_dataset( 'dims/lz'   , data=lz   ) ## [m]
            
            ## Initialize datasets : covariance,Pkz,Pf
            for scalar in scalars:
                hfw.create_dataset( f'covariance/{scalar}' , shape=(nx,)    , dtype=np.float64 , chunks=None    , data=np.full((nx,),0.,np.float64)    )
                hfw.create_dataset( f'Pkz/{scalar}'        , shape=(nx,nkz) , dtype=np.float64 , chunks=(1,nkz) , data=np.full((nx,nkz),0.,np.float64) )
                hfw.create_dataset( f'Pf/{scalar}'         , shape=(nx,nf)  , dtype=np.float64 , chunks=(1,nf)  , data=np.full((nx,nf),0.,np.float64)  )
            
            ## Initialize datasets : 1D [x] mean
            for scalar in avg_Re.dtype.names:
                hfw.create_dataset( f'avg/Re/{scalar}', shape=(nx,), dtype=np.float64, chunks=None, data=np.full((nx,),0.,np.float64) )
    
    self.comm.Barrier()
    
    ## Re-open in parallel for data writes
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## Collectively write covariance,Pkz,Pf
        for scalar in scalars:
            dset = hfw[f'covariance/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = covariance[scalar][:]
            dset = hfw[f'Pkz/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = Pkz[scalar][:,:]
            dset = hfw[f'Pf/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = Pf[scalar][:,:]
        
        ## Collectively write 1D [y] avgs
        for scalar in avg_Re.dtype.names:
            dset = hfw[f'avg/Re/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = avg_Re[scalar][:]
    
    ## Report file contents
    self.comm.Barrier()
    if (self.rank==0):
        even_print( os.path.basename(fn_h5_out) , f'{(os.path.getsize(fn_h5_out)/1024**2):0.1f} [MB]' )
        print(72*'-')
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : spd.calc_turb_cospectrum_wall() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return
